#ifndef _UTIL_H

typedef struct 
{
    char atomname[5],alt, resname[4],chain, reseq[5], icode;
    float x,y,z,occ,b;
}PDBLINE;

#define _UTIL_H
#include "CifFile.h"
#include "TableFile.h"

void token_str(char *str, const char *token, long *nstr,
               char **line);
char ** string_token(char *str, const char *token, int *nstr);
char **token_pattern(char *str_in, const char *tok, int *ntok);
char **token_pattern_case(char *str_in, const char *tok, int *ntok);

long upperstr(char *a);
long lowerstr(char *a);
long num_of_line(const char *inpfile);
void get_data_set_id(const char *str, const char *start, const char *end,
                      char *substr);
void char2space(char *str, char charact);
void rid_of_space(char *str);

void rid_of_char(char *str, char ch);

void rid_of_front_end_space(char *str);

void rid_of_front_space(char *str);

void rid_of_end_space(char *str);
int strstr_space( const char *str, const char *str1);
int pattern5(const char *str,const char *str1,const char *str2,
             const char *str3, const char *str4,const char *str5);
int strcmp_case (const char *s1, const char *s2);
int strncmp_case (const char *s1, const char *s2, int n);
char* strstr_case(const char *s1, const char *s2);

void replace_char(char *str, char find, char replace);
void replace_first_str(const char *str_old, const char *str1, const char *str2,
                       char *str_new);
char *replace_all_str(const char *str_in, const char *str1, const char *str2);

int is_space_line(const char *str);
void change_file_name(const char *inpfile, char *outfile);

void delete_return(char *line);
void get_string(const char *str_in, int start, int end, char *str_out);
void get_last_string(const char *str_in, int length,  char *str_out);
char *get_substr(const char *str,  const char *start,  const char *end);
char *get_string_by_length(const char *str,  int start,   int end);
void get_string_with_length(const char *str, int start,int end, char *substr);
char *get_string_after_id_number(const char *str,  const char *start,  int num);
char *get_string_after_id_length(const char *str,  const char *start,  int len);
char *get_string_patt(int n, int nline, char **line,
                      const char *start, const char *end,int *j);

char *pdb_string_patt(int n, int nline, char **line, int pos,
                      const char *start, int *j);


int get_string_line_patt(int n, int nline, char **line,
                         const char *end1,const char *end2,const char *end3);
float get_float(const char *str, const char *start,  const char *end);
int get_int(const char *str,  const char *start,   const char *end);
float get_float_by_length(const char *str,  int start,   int end); 
int get_int_by_length(const char *str,  int start,   int end); 
void delete_file(const char *file);
char *make_string(const char *str1, const char *str2, const char *str3);
char **get_lines_from_file(const char *inpfile, int *nline); //space line deleted



char *get_attribute_value(CifFile *fobj, const char *blockId,
                          const char *category, const char *attribute);
char **get_attribute_values(CifFile *fobj, const char *blockId,
                            const char *category,const char *attribute,
                            long *nrow);
int  check_category(const char *str, char *category, char *item);
void write_cif(FILE *fout,  const char *inpfile,const  char *method);
int  write_mmcif_sf(FILE *fout, char *iFile, int write);
void array_sort(char ** str, int max);
void larray_sort(int *array, int max, int key);

void print_mmcif_head(FILE *fout, const char *category,
                      const char **items, int nitems);
void write_mmcif_head(FILE *fout, const char *category,const char **items);

void open_file_error(const char *inpfile, const char *message);
void pinfo(char *info, int id);

char **set_memory_2d( int nptr, int length);
void free_memory_2d(char **var, int nptr);
int **set_imemory_2d( int nptr, int length);
void free_imemory_2d(int **var, int nptr);

float **set_fmemory_2d( int nptr, int length);
void free_fmemory_2d(float **var, int nptr);
double **set_dmemory_2d( long nptr, long length);
void free_dmemory_2d(double **var, long nptr);



void file_handle(char *inpfile,char *outfile);
void  get_one_string_after_id(const char *str,  const char *id,  char *str_out);
int get_int_after_id(const char *str,  const char *id);
float get_float_after_id(const char *str,  const char *id);
int space_group_convertor(char *sg_name_inp, char *sg_name_out);

FILE *get_temp_file( char *inpfile);
void rid_of_file_path_unix(char *str);
void clean_file_path(char *str, char *path, char *str_new);


int str_cmp(const char *str, const char *str1, int n1, char *tmp1);
int str_comp(const char *str, const char *str1,  char *tmp1);
int is_protein_na_residue(int nres, char *res, const char **residues30);
int is_protein(char *resname);
int is_nucleic_acid(char *resname);
void dos2unix(char *file);
char * check_pdb(char *inpfile);
void parse_pdbline(char *str);
void clean_string(const char *str1,  char *str2);

void clean_file(char *infile, char *outfile);
int non_ascii(char *infile);
int find_atom(const char *str, char **strmat, int nb, int ne);
void get_poly_type(int n, int i, int **ch_idx, int **res_idx, char **atomname, int **pptype, char **type);
int protein_or_na(int n, int **ch_idx, char **resname, int **res_idx, char **atomname);

int protein_or_na_by_residue(char *resname);
void get_atominfo(int nline, char **line, char **atomname, float **xyz,int **residx, int *nres);

int **chain_idx(int nres, char **chains, int *nch);

void atominfo(int nline, char **line, char **atomname, char **resname,
             char **chain, float **xyz, int **residx, int *nres,
             int **chidx, int *nchain);
void get_nchain_nres(int nline, char **line, int *nres, int *nch);

int is_prime(char **atomname, int n1, int n2);

float dist(float *v1, float *v2);

void check_system_call(int n, const char *idd);
void get_local_file(char *file, const char *file_name);

char *check_env(const char *str);
int check_file(int nchr, const char* filename);
int replace_na_name(char *str, char *str_new);


#endif                                /* _UTIL_H */
