/*  convert a structure factor file from mmCIF to new mmCIF */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include "rcsb_math.h"
#include "sf_convert.h"
#include "util.h"

#include "TableFile.h"
#include "CifFile.h"
#include "CifParserBase.h"


extern char SYMM[20], FREERV[20], PDBID[80], INFO[512],DETAIL[9999];
extern int TEST, SINGLE, VERB, ADD_MAP;
extern float CELL[6], WAVE, RESCUT, SIGCUT, WAVE, RESOH, RESOL;

SF sf;

SF::SF() {
  /* Nothing to do - initialization will zero data */
  //printf("SF INIT\n");
}

SF::~SF() {
  //printf("SF DESTRUCT\n");
  empty();
}

// Empty and clear pointer
void SF::sf_delete_attribute_value(char *&value)
{
  delete_attribute_value(value);
  value = NULL;
}

void SF::sf_delete_attribute_values(char **&value)
{
  delete_attribute_values(value);
  value = NULL;
}
  
void SF::empty() {
  //printf("SF EMPTY\n");
  sf_delete_attribute_value(a);
  sf_delete_attribute_value(b);
  sf_delete_attribute_value(c);
  sf_delete_attribute_value(alpha);
  sf_delete_attribute_value(beta);
  sf_delete_attribute_value(gamma);
  sf_delete_attribute_value(sg);
  sf_delete_attribute_value(nsg);
  sf_delete_attribute_value(resh);
  sf_delete_attribute_value(resl);
  sf_delete_attribute_value(hmax);
  sf_delete_attribute_value(kmax);
  sf_delete_attribute_value(lmax);
  sf_delete_attribute_value(hmin);
  sf_delete_attribute_value(kmin);
  sf_delete_attribute_value(lmin);
  sf_delete_attribute_value(nall);
  sf_delete_attribute_value(nobs);
  sf_delete_attribute_values(wavelength);
  sf_delete_attribute_values(symm);
  sf_delete_attribute_values(aud_id);
  sf_delete_attribute_values(aud_date);
  sf_delete_attribute_values(aud_rec);
  sf_delete_attribute_values(aud_meth);
  sf_delete_attribute_values(dif_id);
  sf_delete_attribute_values(temp);
  sf_delete_attribute_values(treat);
  sf_delete_attribute_values(dif_detail);
  sf_delete_attribute_values(H);
  sf_delete_attribute_values(K);
  sf_delete_attribute_values(L);
  sf_delete_attribute_values(Fo);
  sf_delete_attribute_values(sFo);
  sf_delete_attribute_values(Fo_au);
  sf_delete_attribute_values(sFo_au);
  sf_delete_attribute_values(Fc);
  sf_delete_attribute_values(Fc_au);
  sf_delete_attribute_values(Io);
  sf_delete_attribute_values(sIo);
  sf_delete_attribute_values(F2o);
  sf_delete_attribute_values(sF2o);
  sf_delete_attribute_values(Ic);
  sf_delete_attribute_values(F2c);
  sf_delete_attribute_values(fom);
  sf_delete_attribute_values(phase_c);
  sf_delete_attribute_values(phase_o);
  sf_delete_attribute_values(status);
  sf_delete_attribute_values(flag);
  sf_delete_attribute_values(free1);
  sf_delete_attribute_values(free2);
  sf_delete_attribute_values(wave_id);
  sf_delete_attribute_values(cryst_id);
  sf_delete_attribute_values(scale_id);
  sf_delete_attribute_values(I_plus);
  sf_delete_attribute_values(sI_plus);
  sf_delete_attribute_values(I_minus);
  sf_delete_attribute_values(sI_minus);
  sf_delete_attribute_values(F_plus);
  sf_delete_attribute_values(sF_plus);
  sf_delete_attribute_values(F_minus);
  sf_delete_attribute_values(sF_minus);
  sf_delete_attribute_values(hla);
  sf_delete_attribute_values(hlb);
  sf_delete_attribute_values(hlc);
  sf_delete_attribute_values(hld);
  sf_delete_attribute_values(hla1);
  sf_delete_attribute_values(hlb1);
  sf_delete_attribute_values(hlc1);
  sf_delete_attribute_values(hld1);
  sf_delete_attribute_values(anom);
  sf_delete_attribute_values(anoms);
  sf_delete_attribute_values(d_spacing);
  sf_delete_attribute_values(a_calc);
  sf_delete_attribute_values(b_calc);
  sf_delete_attribute_values(weighted);
  sf_delete_attribute_values(fs_uncorrected);
  sf_delete_attribute_values(f_uncorrected);
  sf_delete_attribute_values(gsas_i100_meas);
  sf_delete_attribute_values(i_unknown);
  sf_delete_attribute_values(is_unknown);
  sf_delete_attribute_values(phase_part);
  sf_delete_attribute_values(F_part_au);
  sf_delete_attribute_values(Fc_all);
  sf_delete_attribute_values(Fc_all_ph);
  sf_delete_attribute_values(fwt);
  sf_delete_attribute_values(phwt);
  sf_delete_attribute_values(delfwt);
  sf_delete_attribute_values(delphwt);
  sf_delete_attribute_values(fib_co);
  sf_delete_attribute_values(fib_fo);
  sf_delete_attribute_values(fib_ly);
  sf_delete_attribute_values(wt);
  sf_delete_attribute_values(pow_sc);
  sf_delete_attribute_values(pow_io);
  sf_delete_attribute_values(pow_iop);
  sf_delete_attribute_values(pow_wt);
  sf_delete_attribute_values(pow_ic);
  sf_delete_attribute_values(pow_ict);
  sf_delete_attribute_values(dH);
  sf_delete_attribute_values(dK);
  sf_delete_attribute_values(dL);
  sf_delete_attribute_values(unmerge_i);
  sf_delete_attribute_values(unmerge_si);
}

void cif2cif_sf_all(char *iFile, char *sffile_new, char *pdb_id, int key)
//parse cif and rewrite.
//key=0, only write 1st block;
//key=1, parse data and write statistics
//key=2,
    
{
    string diags, blockId;
    char fname[2000], **line, block_id[256];
    int  i=0, numBlocks, nstr,  c;
    FILE *fp, *fout;
    
    CifFile * fobjR = NULL;
    CifParser* cifParserR = NULL;

    if (!check_file(10, iFile)) {
        sprintf(INFO, "Error: Input file (%s) size is too small! No conversion performed.", iFile);
        pinfo(INFO,0);
        return;
    }
    
  //  printf("\nDoing mmcif to mmcif conversion..\n");

    strcpy(fname, "");
    
    if((fout = fopen(sffile_new, "w"))==NULL){
        open_file_error(sffile_new, "cif2cif_sf_all");
    }
    
    fobjR = new CifFile(true);
    cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
    cifParserR->Parse(iFile, diags);
    delete(cifParserR);

    if (!diags.empty()) {
        sprintf(INFO, "%s\n",diags.c_str());
        pinfo(INFO,0);
        if (strstr(INFO, "values is not exact multiples of the number")){
            sprintf(INFO, "Error: Conversion stopped!\n");
            pinfo(INFO,0);
            return;
        }
    }

    vector<string> blockNames;
    fobjR->GetBlockNames(blockNames);
    blockId = fobjR->GetFirstBlockName();
    numBlocks = fobjR->GetNumBlocks();
        
    sprintf(INFO, "Total number of data blocks = %d\n", numBlocks);
    if (key==1) pinfo(INFO,0);
    for(i=0; i<numBlocks;i++){ 
        blockId = blockNames[i];
        strcpy(block_id, blockId.c_str());
        if (!strcmp_case(pdb_id, "xxxx")){
            if (strlen(block_id)>6){
                strncpy(pdb_id, block_id+1, 4);
                pdb_id[4]='\0';
		// Avoid strcpy for overlapping memory
		if (PDBID != pdb_id) {
		  strcpy(PDBID, pdb_id);
		}
            }
        }
        
        sprintf(INFO, "data_block_id=%s,  pdbid=%s,  block_number=%d\n", block_id, pdb_id, i+1);
        if (key==1) pinfo(INFO,0);
        write_hkl_values(fout, fname, fobjR, i, numBlocks, pdb_id, key);
        if(key==0 && i==0) { //only first block and give F if it is not there.
            fclose(fout);
            if (fobjR) delete fobjR;
            return;
        }
    }
    if (fobjR) delete fobjR;
    
    line = string_token(fname, "," , &nstr);
//    for(i=0; i<nstr; i++){ printf("FILE=%d || %s\n", i+1,line[i]); }
//    column=0;
//    array_sort_by_column(line, column, 2, nstr);

    rewind(fout);
    for(i=0; i<nstr; i++){
        sscanf(line[i], "%*s%s",fname);
        
        fp=fopen(fname, "r");
        while((c=getc(fp))!=EOF){ // write to given file name
            fprintf(fout,"%c",c);
        }
        fclose(fp);
        
        if (VERB==0) delete_file(fname);
    }
    fclose(fout);
    if(nstr) free_memory_2d(line,nstr);
}

void sf_stat(char *inpfile)
//parse cif and get statistics.
    
{
    string diags, blockId;
    char  block_id[256];
    int  i=0, numBlocks;
    float resol;
    
    CifFile * fobjR = NULL;
    CifParser* cifParserR = NULL;

    if (!check_file(10, inpfile)) {
        sprintf(INFO, "Error: Input file (%s) size is too small! No conversion performed.", inpfile);
        pinfo(INFO,0);
        return;
    }
    
    printf("\nGeting statistics from (%s)\n", inpfile);
    
    fobjR = new CifFile(true);
    cifParserR = new CifParser(fobjR, fobjR->GetVerbose());
    cifParserR->Parse(inpfile, diags);
    delete(cifParserR);

    if (!diags.empty()) {
        sprintf(INFO, "%s\n",diags.c_str());
        pinfo(INFO,0);
        if (strstr(INFO, "values is not exact multiples of the number")){
            sprintf(INFO, "Error: Conversion stopped!\n");
            pinfo(INFO,0);
            return;
        }
    }

    vector<string> blockNames;
    fobjR->GetBlockNames(blockNames);
    blockId = fobjR->GetFirstBlockName();
    numBlocks = fobjR->GetNumBlocks();
        
    sprintf(INFO, "Total number of data blocks = %d\n", numBlocks);
    pinfo(INFO,0);
    for(i=0; i<numBlocks;i++){ 
        blockId = blockNames[i];
        strcpy(block_id, blockId.c_str());
        
        parse_sf_values(fobjR, i);
        sprintf(INFO, "\nData_block_id=%s,  pdbid=%s,  block_number=%d\n", block_id, PDBID, i+1);
        pinfo(INFO,0);
        check_sf(i, &resol) ;
        if(i==0 && sf.nref>50) write_sf_4_validation(); //first block for validation

    }
    if (fobjR) delete fobjR;
}


void i_to_f(int i, char *I , char **sI, char* fp, char *sigfp, float sf_defaut)
{
    float f=0, sf=0;

    if(atof(I)>0){
        f=sqrt(atof(I));
        if(sI && atof(sI[i])>0){
            sf=atof(sI[i])/(2.0*f);
        }else
            sf=sf_defaut;

    }
    sprintf(fp, "%.2f" , f);
    sprintf(sigfp, "%.2f" , sf);
}

void other_to_f(int i, float &F, float &Fs)
// key=0,1
{
    float sp1=0, sp2=0, I=0, Is=0;
    
    F = 0;
    Fs = 0.01;
    
    if(sf.Io){ // Io
        if(!strchr(sf.Io[i], '?')){
            I=atof(sf.Io[i]);
            
            if( I>0 ){
                F=sqrt(I);
                if(sf.sIo) Fs= atof(sf.sIo[i])/(2.0*F);
            }
        }
        
    }else if(sf.sI_plus && sf.sI_minus){ // I_plus
        sp1=1;
        sp2=1;
        (strchr(sf.sI_plus[i],'?'))? sp1=-1 : sp1=atof(sf.sI_plus[i]);
        (strchr(sf.sI_minus[i],'?')) ? sp2=-1 : sp2=atof(sf.sI_minus[i]);
            
        I = 0;
        Is=0;
        if(sp1<0 && sp2<0){
            I = 0;
            Is=0;
        }else if(sp1>0 && sp2<0){
            I=atof(sf.I_plus[i]);
            Is=atof(sf.sI_plus[i]);
            
        }else if(sp1<0 && sp2>0){
            I=atof(sf.I_minus[i]);
            Is=atof(sf.sI_minus[i]);
        
        }else if(sp1>0 && sp2>0){
            I = (atof(sf.I_plus[i])+atof(sf.I_minus[i]))/2.0;
            Is= (atof(sf.sI_plus[i])+atof(sf.sI_minus[i]))/2.0;
        }
        if(I>0) {
            F=sqrt(I);
            Fs=Is/(2.0*F);
        }
        
    }else if(sf.F_plus && sf.F_minus){ // F_plus
        
        if(strchr(sf.F_plus[i], '?') && strchr(sf.F_minus[i], '?')){
            F = 0;
            Fs=0;
        }else if(!strchr(sf.F_plus[i], '?') && strchr(sf.F_minus[i], '?')){
            F=atof(sf.F_plus[i]);
            Fs=atof(sf.sF_plus[i]);
        }else if(strchr(sf.F_plus[i], '?') && !strchr(sf.F_minus[i], '?')){
            F=atof(sf.F_minus[i]);
            Fs=atof(sf.sF_minus[i]);
        
        }else if(!strchr(sf.F_plus[i], '?') && !strchr(sf.F_minus[i], '?')){
            F=(atof(sf.F_plus[i])+atof(sf.F_minus[i]))/2.0;
            Fs=(atof(sf.sF_plus[i])+atof(sf.sF_minus[i]))/2.0;
        }
    }
    
}

void write_sf_4_validation()
// write a sf file for validation. Add F/sF if missing.
{
    char fp[20], sigfp[20], flag;
    int n=0,nf=0, i, ah,ak, al;
    float F=0, Fs=0,sig=0.01, cell[6], rcell[6], resol, i_sigi, af,afs;
    FILE *fout=NULL;

    n=sf.nref;


    remove("SF_4_validate.cif");
    fout=fopen("SF_4_validate.cif", "w"); //auto-refine use this name

    fprintf(fout, "data_cif2cif\n#\n");

    if (CELL[0]>2 && CELL[4]>2){
        cryst_infor(fout, CELL, SYMM , 0);
        calc_recip_cell(CELL, rcell);
    }else if(sf.a &&  sf.alpha && sf.beta  && sf.sg){
        cell[0] = atof(sf.a);
        cell[1] = atof(sf.b);
        cell[2] = atof(sf.c);
        cell[3] = atof(sf.alpha);
        cell[4] = atof(sf.beta);
        cell[5] = atof(sf.gamma);
        
        cryst_infor(fout, cell, sf.sg , 0);        
        calc_recip_cell(cell, rcell);        
    }

    fprintf(fout, "\n#\n");
    fprintf(fout, "loop_\n");
    fprintf(fout, "_refln.index_h\n");
    fprintf(fout, "_refln.index_k\n");
    fprintf(fout, "_refln.index_l\n");
    fprintf(fout, "_refln.status\n" );
    fprintf(fout, "_refln.F_meas_au\n");
    fprintf(fout, "_refln.F_meas_sigma_au\n");
    fprintf(fout, "_refln.d_spacing\n"); 
    fprintf(fout, "_refln.I_over_sigI\n");
    
    if (sf.Io && sf.sIo){
        fprintf(fout, "_refln.intensity_meas\n");
        fprintf(fout, "_refln.intensity_sigma\n");
    }
    
    if (sf.F_plus && sf.sF_plus && sf.F_minus && sf.sF_minus){
        fprintf(fout, "_refln.pdbx_F_plus \n" );
        fprintf(fout, "_refln.pdbx_F_plus_sigma \n" );
        fprintf(fout, "_refln.pdbx_F_minus \n" );
        fprintf(fout, "_refln.pdbx_F_minus_sigma \n" );
    }
    
    if (sf.I_plus && sf.sI_plus &&  sf.I_minus && sf.sI_minus){
        fprintf(fout, "_refln.pdbx_I_plus \n" );
        fprintf(fout, "_refln.pdbx_I_plus_sigma \n" );
        fprintf(fout, "_refln.pdbx_I_minus \n" );
        fprintf(fout, "_refln.pdbx_I_minus_sigma \n" );
    }

    strcpy(fp, "?");
    strcpy(sigfp, "?");    
    for(i=0; i<n; i++){
        if(sf.status){            
            if (sf.status[i][0] != 'o' &&  sf.status[i][0] != 'f' ) continue;
            flag=sf.status[i][0];
        }else{
            flag='o';
        }
       
        if(sf.Fo_au ){
            strcpy(fp, sf.Fo_au[i]);
            if(sf.sFo_au){
                if (atof(sf.sFo_au[i])>0 && sf.sFo_au[i][0] !='?')
                   strcpy(sigfp,sf.sFo_au[i]);
                else
                   sprintf(sigfp,"%.4f", sig);
            }else{
                sprintf(sigfp,"%.4f", sig);
            }

        }else if(sf.Io ){
            i_to_f(i, sf.Io[i], sf.sIo, fp, sigfp, sig);

        }else if(sf.I_plus || sf.F_plus){
            other_to_f(i, F, Fs) ;
            sprintf(fp,"%.4f", F);
            sprintf(sigfp,"%.4f", Fs);
            
        }else if(sf.Fo){
            strcpy(fp, sf.Fo[i]);
            if(sf.sFo){
                if (atof(sf.sFo[i])>sig)
                   strcpy(sigfp,sf.sFo[i]);
                else
                   sprintf(sigfp,"%.4f", sig);
            }else{
                sprintf(sigfp,"%.4f", sig);
            }
        }else if(sf.F2o){
            i_to_f(i, sf.F2o[i], sf.sF2o, fp, sigfp, sig);  
        }
        
        ah = atoi(sf.H[i]);
        ak = atoi(sf.K[i]);
        al = atoi(sf.L[i]);
        resol=get_resolution(ah, ak, al, rcell);
        
        i_sigi=0;
        if (sf.Io && sf.sIo && atof(sf.sIo[i])>0){
            i_sigi=atof(sf.Io[i])/atof(sf.sIo[i]);
        }else{
            af=atof(fp);
            afs=atof(sigfp);
            if (afs>0) {i_sigi=af/(2*afs);}
        }
        
        if(RESCUT>0.0001 && resol<RESCUT && resol>0.0001) continue;
        if(fabs(SIGCUT)>0.0001 && i_sigi<SIGCUT ) continue;
        
        fprintf(fout, " %4s %4s %4s %c %8s %8s ", sf.H[i], sf.K[i], sf.L[i], flag, fp, sigfp);
        fprintf(fout, "%6.2f %6.2f", resol, i_sigi);

        if (sf.Io && sf.sIo){
            fprintf(fout, " %4s ",sf.Io[i]);
            (sf.sIo)? fprintf(fout, " %4s ",sf.sIo[i]): fprintf(fout, " ? ");
        }
        
        if (sf.F_plus && sf.sF_plus && sf.F_minus && sf.sF_minus){
            fprintf(fout, "  %4s %4s %4s %4s ", sf.F_plus[i], sf.sF_plus[i], sf.F_minus[i],sf.sF_minus[i]);
        }
        
        if (sf.I_plus && sf.sI_plus &&  sf.I_minus && sf.sI_minus){
            fprintf(fout, "  %4s %4s %4s %4s ", sf.I_plus[i], sf.sI_plus[i], sf.I_minus[i],sf.sI_minus[i]);
        }
        
        fprintf(fout,"\n");
        nf++;

    }
    fclose(fout);

    sprintf(INFO,"\nNumber of reflections for validation set =%d",  nf);
    pinfo(INFO,0);
    
}

void write_sf_tail_identifer(FILE *fo, int i, int nd)
{
    if(nd==1)
        fprintf(fo,"#END OF REFLECTIONS\n");
    else{
        if(i<nd-1)
            fprintf(fo,"#END\n");
        else
            fprintf(fo,"#END OF REFLECTIONS\n");
    }
}

int check_hkl_duplicate(int nref, int block, char **hkl)
{
    int j, ndup=0;
    
    if(nref>0){
        quicksort(hkl,nref);
        for (j=1; j<nref; j++){
            
            if (!strcmp(hkl[j], hkl[j-1]) ){
                if (ndup++ >4 ) continue;
                
                sprintf(INFO, "Warning: Duplicated H,K,L (%s) (data block=%d).", hkl[j], block+1);
                pinfo(INFO,0);
                    
            }
        }
    }
    return ndup;
}

void check_cell(float *cell, int nblock)
//check cell consistence and assign it to cell
{
    
    if(sf.a && sf.b && sf.c && sf.alpha && sf.beta && sf.gamma ){  
        cell[0] = atof(sf.a);
        cell[1] = atof(sf.b);
        cell[2] = atof(sf.c);
        cell[3] = atof(sf.alpha);
        cell[4] = atof(sf.beta);
        cell[5] = atof(sf.gamma);

        if(CELL[0]>0.01 && nblock==0 &&(
            fabs(cell[0] - CELL[0])>0.1 || fabs(cell[1] - CELL[1])>0.1 || fabs(cell[2] - CELL[2])>0.1 ||
            fabs(cell[3] - CELL[3])>0.1 || fabs(cell[4] - CELL[4])>0.1 || fabs(cell[5] - CELL[5])>0.1) ){
            sprintf(INFO,"Warning! SF and PDB (%s nblock=%d) cell values mismatch.", PDBID, nblock+1);
            pinfo(INFO, 0);

            if (fabs(cell[0] - CELL[0])>3.0 || fabs(cell[1] - CELL[1])>3.0 || fabs(cell[2] - CELL[2])>3.0 ||
                fabs(cell[3] - CELL[3])>3.0 || fabs(cell[4] - CELL[4])>3.0 || fabs(cell[5] - CELL[5])>3.0 ){
              
                sprintf(INFO,"Error: (%s nblock=%d) large cell value mismatch (>3.0).", PDBID, nblock+1);
                printf("cell in  sf: %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f\n",
                       cell[0],cell[1], cell[2],cell[3],cell[4],cell[5]);
                printf("cell in PDB: %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f\n",
                       CELL[0],CELL[1], CELL[2],CELL[3],CELL[4],CELL[5]);
                
                pinfo(INFO, 0);
            }
            
        }
    }else if(CELL[0]>0.01 && CELL[1]>0.01 && CELL[2]>0.01 &&
             CELL[3]>0.01 && CELL[4]>0.01 && CELL[5]>0.01 ){  
        cell[0] = CELL[0];
        cell[1] = CELL[1];
        cell[2] = CELL[2];
        cell[3] = CELL[3];
        cell[4] = CELL[4];
        cell[5] = CELL[5];
    }
}

void write_hkl_values(FILE *fout, char *fname, CifFile *fobjR, int i,int nd, char *pdb_id, int key)
// i= data block number ; nd=total number of blocks; key=0,1,2 a contral number.
{
    char outfile[512], block_name[256],blockId[80],tmp1[500],bid[80], **hkl;
    vector<string> blockNames;
    int index[80],nf=0, j, nblock=0, ndup=0;
    float resol=50;
    FILE *fo=NULL;
    
    strcpy(bid, "_ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789abcdefghijklmnopqrstuvwxyz");
    nblock=i;
    
    fobjR->GetBlockNames(blockNames);
    
    strcpy(blockId , blockNames[i].c_str());
    rid_of_space(blockId);
    if(strchr(blockId, '/')) replace_char(blockId, '/', '_');
    if(strchr(blockId, ')')) replace_char(blockId, ')', '_');
    if(strchr(blockId, '(')) replace_char(blockId, '(', '_');

    if(strlen(blockId)==0){ strcpy(blockId, "UNK"); }
    
    parse_sf_values(fobjR, i);

    hkl = set_memory_2d (sf.nref, 20);
    
    for (j=0; j<sf.nref; j++){//merge two free set to one!(mainly phenix I+,I-)
           
        if (sf.status && sf.free1){
            if (sf.status[j][0]=='?' && sf.free1[j][0]!='?'){
                strcpy(sf.status[j], sf.free1[j]);
            }
            
        }else if (sf.status && sf.free2 ){
            if(sf.status[j][0]=='?' && sf.free2[j][0]!='?'){
                strcpy(sf.status[j], sf.free2[j]);
            }
        }
        
        if (i>=0){ //test duplicate of HKL
            sprintf(hkl[j], "%s %s %s", sf.H[j],sf.K[j],sf.L[j]);
        }
    }
    
    if(sf.nref>0) ndup = check_hkl_duplicate(sf.nref, i, hkl);
    
    if ((ndup>1 && (sf.Io || sf.I_plus )) || sf.unmerge_i){ //unmerged data
        if (i==0){
            sprintf(INFO, "Warning: Unmerged data in block 1 (pdbid=%s; blockId=%s)!", pdb_id,blockId);
        }else{
            sprintf(INFO, "Note: Unmerged data in (pdbid=%s; blockId=%s)!", pdb_id,blockId);
        }
        pinfo(INFO,0);
        
        sprintf(block_name, "r%s%csf",pdb_id,bid[i]);
        sprintf(outfile, "SF_%s_%c.mmcif",pdb_id,bid[i]);
//        sprintf(outfile, "SF_%s.mmcif",blockId);
        if((fo = fopen(outfile, "w"))==NULL){        
            printf("Error! No permision to write the file %s\n",outfile);
            sprintf(tmp1,"%.3f %s,", resol,outfile);
            strcat(fname, tmp1);
	    free_memory_2d(hkl, sf.nref);
            return ;
        }
        if(i>0){ //do not change token for 1st block even it is unmerged.
            write_sf_unmerge(fo, block_name, pdb_id, i, index, key);
            write_sf_tail_identifer(fo,i,nd);
            fclose(fo);
            sprintf(tmp1,"%.3f %s,", resol,outfile);
            strcat(fname, tmp1);
	    free_memory_2d(hkl, sf.nref);
            return;
        }
        
    }
    
    if(sf.nref<=10 && sf.dnref<=10 ) {
        sprintf(INFO, "Error: Number of reflections <10 (pdbid=%s; blockId=%s)!", pdb_id, blockId);
        pinfo(INFO,0);
	free_memory_2d(hkl, sf.nref);
        return;
    }

    get_num_file(sf.cryst_id, sf.wave_id, sf.scale_id, sf.nref, index, &nf) ;
    nf=1;
    
    if(key==0 && i==0){
        sprintf(block_name, "r%ssf",pdb_id);
        sf_related_head(fout, block_name, pdb_id, i);
        write_sf_file(fout, block_name, 1, i, index, key);
	free_memory_2d(hkl, sf.nref);
        return;
    }
    
    if (key==1) check_sf(i, &resol);
    if(i==0 && sf.status){ check_sf_status(sf.status, i, PDBID); }

    if(SINGLE==0){//block by block (current method)
        sprintf(outfile, "SF_TEMP_%d.mmcif",nblock);
        if((fo = fopen(outfile, "w"))==NULL){        
            printf("Error: I have no permision to write the file %s\n",outfile);
            sprintf(tmp1,"%.3f %s,", resol,outfile);
            strcat(fname, tmp1);
	    free_memory_2d(hkl, sf.nref);
            return ;
        }

        if(i==0){ //first block
            sprintf(block_name, "r%ssf",pdb_id);
        }else{ //other block
            sprintf(block_name, "r%s%csf",pdb_id,bid[i]);
        }

        sf_related_head(fo, block_name, pdb_id, i);
        write_sf_file(fo, block_name, 1, i, index, key);
        write_sf_tail_identifer(fo,i,nd);
        fclose(fo);
        
        sprintf(tmp1,"%.3f %s,", resol,outfile);
        strcat(fname, tmp1);
        
    }else{ // parse all blocks
        for(j=0; j<nf; j++){
            
            if(j==0){
                if(strlen(pdb_id)>0){
                    sprintf(block_name, "%s_%d",pdb_id,j);
                    sf_related_head(fo, block_name, pdb_id, i);
                    
                }else{
                    sprintf(block_name, "%s_%d",blockId,j);
                    fprintf(fo, "data_%s\n",block_name);
                }
            }else{
                sprintf(block_name, "%s_%d",blockId,j);
                fprintf(fo, "data_%s\n",block_name);
            }
            
            
            if(sf.cryst_id && sf.wave_id){
                printf("SF extracting (Crystal ID = %s ;  Wavelength ID = %s):\n",
                       sf.cryst_id[index[j]], sf.wave_id[index[j]]);
            
                sprintf(outfile, "SF_%s_%s_%s.cif",
                        sf.cryst_id[index[j]], sf.wave_id[index[j]], blockId);
            }else if(sf.cryst_id && !sf.wave_id){    
                printf("SF extracting (Crystal ID = %s):\n", sf.cryst_id[ index[j] ]);             
                sprintf(outfile, "SF_%s_%s_1.mmcif",blockId, sf.cryst_id[index[j]]);
                
            }else if(sf.wave_id && !sf.cryst_id){    
                printf("SF extracting (Wavelength ID = %s):\n", sf.wave_id[ index[j] ]);
                sprintf(outfile, "SF_%s_1_%s.mmcif",blockId, sf.wave_id[index[j]]);
            
            }else
                sprintf(outfile, "SF_%s_1_1.mmcif",blockId);
            
        
            if((fo = fopen(outfile, "w"))==NULL){        
                printf("Error: No permision to write the file %s\n",outfile);
                return ;
            }
            
            write_sf_file(fo, block_name, 2, j, index, key);

            fprintf(fo,"#END OF REFLECTIONS\n");
            
            fclose(fo);
            printf("The output file=%s %.2f\n",outfile, resol);
        
            sprintf(tmp1,"%.3f %s,", resol,outfile);
            strcat(fname, tmp1);
        }
    }
    free_memory_2d(hkl, sf.nref);
}

void get_num_file(char **cryst_id, char **wave_id, char **scale_id,
                  int n, int *index, int *nf)
// n is # of reflections for the data-block.
// index  corresponding to HKL number (index[0]=1; index[1]=n)
{
    int i=0,  m1=0;

    if(!cryst_id){
        cryst_id = new char *[n+1];
        for(i=0; i<n; i++){
            cryst_id[i]=new char[2];
            strcpy(cryst_id[i], "1");
        }
    }
    
    if(!wave_id){
        wave_id = new char *[n+1];
        for(i=0; i<n; i++){
            wave_id[i]=new char[2];
            strcpy(wave_id[i], "1");
        }
    }
    
    if(!scale_id){
        scale_id = new char *[n+1];
        for(i=0; i<n; i++){
            scale_id[i]=new char[2];
            strcpy(scale_id[i], "1");
        }
    }
    num_file(cryst_id, wave_id, scale_id, n, index, &m1) ;
    *nf=m1;
}

void num_file(char **cryst_id, char **wave_id,char **scale_id, int n, int *index, int *nf) 
//get total different reflection files (nf,index)
{
    long i=0,  m1;
    char  **key=NULL, tmp[300], tmp1[100], tmp2[100], tmp3[100];


    key = new char *[n+1];
    
    (cryst_id) ? strcpy(tmp1,cryst_id[0]) :  strcpy(tmp1,"");
    (wave_id) ? strcpy(tmp2,wave_id[0]) :  strcpy(tmp2,"");
    (scale_id) ? strcpy(tmp3,scale_id[0]) :  strcpy(tmp3,"");
    sprintf(tmp,"%s_%s_%s", tmp1, tmp2, tmp3);
    
    key[0] = new char[strlen(tmp)+1];
    strcpy(key[0], tmp);

    m1=0;
    index[0]=0;
    for(i=1; i<n; i++){
        (cryst_id) ? strcpy(tmp1,cryst_id[i]) :  strcpy(tmp1,"");
        (wave_id) ? strcpy(tmp2,wave_id[i]) :  strcpy(tmp2,"");
        (scale_id) ? strcpy(tmp3,scale_id[i]) :  strcpy(tmp3,"");
        
        sprintf(tmp,"%s_%s_%s", tmp1, tmp2, tmp3);
        
        key[i] = new char[strlen(tmp)+1];
        strcpy(key[i], tmp);

        
        if(strcmp(tmp, key[i-1])){
            m1++;
            index[m1] = i;
        }
    }
    m1++;  // the last one
    index[m1] = n;
    
    *nf=m1;
    if(key)free_memory_2d(key, n);
}

void write_wavelength(FILE *fout, char *pdbid, int block)
{
    int i;
    
    fprintf(fout, "#\nloop_\n");
    fprintf(fout, "_diffrn_radiation_wavelength.id \n");
    fprintf(fout, "_diffrn_radiation_wavelength.wavelength \n");

    if(WAVE>0) {
        fprintf(fout,"1  %.4f \n",  WAVE);
    }else if(sf.nwave){
        for(i=0; i<sf.nwave; i++){
            fprintf(fout,"%d  %s \n", i+1,  sf.wavelength[i]);  
        } 
    }else{
        sprintf(INFO, "Warning: No wavelength value was found in SF file (data block=%d).\n", block);
        pinfo(INFO, 0);
        fprintf(fout,"1  . \n");
    }
}
 
void sf_related_head(FILE *fout_in, char *block_name, char *pdbid, int nd)
/* write all but not _refln */
{
    char sfsg[256],tmp[256], tmp1[256], tmpname[256], **line, date[80], correct[512];
    int i, nline;
    FILE *fout;
    
    
    sprintf(tmpname, "data_%s_tmp.cif", block_name);
    fout=fopen(tmpname, "w");

    fprintf(fout, "data_%s\n",block_name);
    fprintf(fout,"#\n_entry.id  '%s'\n", pdbid);
    fprintf(fout,"#\n_exptl_crystal.id   1 \n");
    fprintf(fout,"#\n_reflns_scale.group_code   1 \n");

    strcpy(date,"");
    strcpy(correct,"");

        
    if (nd==0){
        write_audit(fout, date, correct, sf.naud, sf.aud_id, sf.aud_rec,
                    sf.aud_date, sf.aud_meth, 0);
    }
    if(sf.dif_cid ||  sf.dif_detail || strlen(DETAIL)>0 ||  sf.temp || sf.treat){
      strcpy(tmp,"");
      write_diffrn(fout, sf.ndif, sf.dif_id, sf.dif_cid, sf.dif_detail, sf.temp, sf.treat, tmp);
      
    }
    

    if(sf.a &&  sf.b &&  sf.c && sf.alpha && sf.beta && sf.gamma){  
        fprintf(fout, "#\n_cell.entry_id  %s\n", pdbid);
        fprintf(fout, "_cell.length_a  %s\n", sf.a);
        fprintf(fout, "_cell.length_b  %s\n", sf.b);
        fprintf(fout, "_cell.length_c  %s\n", sf.c);
        fprintf(fout, "_cell.angle_alpha  %s\n", sf.alpha);
        fprintf(fout, "_cell.angle_beta  %s\n", sf.beta);
        fprintf(fout, "_cell.angle_gamma  %s\n", sf.gamma);
        
//        delete []sf.a;delete []sf.b;delete []sf.c;
//        delete []sf.alpha;delete []sf.beta;delete []sf.gamma;
    }

    if(sf.sg) {
        space_group_convertor(sf.sg, sfsg);
        
        strcpy(tmp, sfsg);
        rid_of_front_end_space(tmp);
        
        if(strlen(SYMM)>1){
            space_group_convertor(SYMM, tmp1);
            
            rid_of_front_end_space(tmp1);
            if (strcmp_case(tmp, tmp1)){
                if(nd==0){
                    sprintf(INFO, "Warning! (%s nblock=%d) space group mismatch (pdb= %s : sf= %s).",pdbid,nd+1, tmp1, sfsg);
                }
                pinfo(INFO, 0);
            }
        }
        
        fprintf(fout, "#\n_symmetry.entry_id  '%s'\n",pdbid);
        fprintf(fout, "_symmetry.space_group_name_H-M  '%s'\n",sfsg);
        
        if(sf.nsg)  {
            if(strlen(SYMM)<1 || !strcmp(tmp, tmp1))fprintf(fout, "_symmetry.Int_Tables_number  %s\n", sf.nsg);
        }
    }
    
    if(sf.nsymm){
        fprintf(fout, "#\nloop_\n");
        fprintf(fout, "_symmetry_equiv.id\n");
        fprintf(fout, "_symmetry_equiv.pos_as_xyz\n");
        for(i=0; i<sf.nsymm; i++){
            fprintf(fout,"%d  '%s'\n", i+1,  sf.symm[i]);
        }
    }

        
    if(sf.nwave){
        strcpy(tmp, sf.wavelength[0]);
        if (strcmp(tmp, ".") && (atof(tmp)>2.0 || atof(tmp)<0.6 )){
            sprintf(INFO, "Warning: (%s nblock=%d) wavelength value (%s) is abnormal (double check)!", pdbid,nd+1,tmp);
            pinfo(INFO, 0);
        }
        
        if(WAVE>0.8 && WAVE<1.8 && WAVE !=1.0 && sf.nwave==1){
            if(fabs(WAVE - atof(tmp))>0.0001 && nd==0){
                sprintf(INFO, "Warning: (%s nblock=%d) wavelength mismatch (pdb= %.4f : sf= %s)!",pdbid,nd+1, WAVE,tmp);
                pinfo(INFO, 0);
                sprintf(sf.wavelength[0], "%.4f" , WAVE);
            }
                
        }else if (WAVE>0 && fabs(WAVE - atof(tmp))>0.0001  && nd==0){
            sprintf(INFO, "Warning: (%s nblock=%d) wavelength mismatch (pdb= %.4f : sf= %s). (double check!)",pdbid,nd+1, WAVE,tmp);
            pinfo(INFO, 0);
        }
        
    }    
    
    write_wavelength(fout,pdbid, nd+1);
    
    
    if(sf.resh or sf.nall or sf.nobs ){
        if (!sf.ndif){
            fprintf(fout, "#\n_diffrn.crystal_id   1 \n");
            fprintf(fout, "_diffrn.id   1 \n#\n");
        }
        
        fprintf(fout,"#\n_diffrn_reflns.diffrn_id  1\n");
        if(sf.resh)fprintf(fout,"_diffrn_reflns.pdbx_d_res_high  %s\n", sf.resh);
        if(sf.resl)fprintf(fout,"_diffrn_reflns.pdbx_d_res_low  %s\n", sf.resl);
        if(sf.hmax)fprintf(fout,"_diffrn_reflns.limit_h_max  %s\n", sf.hmax);
        if(sf.hmin)fprintf(fout,"_diffrn_reflns.limit_h_min  %s\n", sf.hmin);
        if(sf.kmax)fprintf(fout,"_diffrn_reflns.limit_k_max  %s\n", sf.kmax);
        if(sf.kmin)fprintf(fout,"_diffrn_reflns.limit_k_min  %s\n", sf.kmin);
        if(sf.lmax)fprintf(fout,"_diffrn_reflns.limit_l_max  %s\n", sf.lmax);
        if(sf.lmin)fprintf(fout,"_diffrn_reflns.limit_l_min  %s\n", sf.lmin);
        if(sf.nall)fprintf(fout,"_diffrn_reflns.number  %s\n", sf.nall);
        if(sf.nobs)fprintf(fout,"_diffrn_reflns.pdbx_number_obs  %s\n", sf.nobs);
    }

    
    fclose(fout);
    
    cifparse(tmpname, tmpname, 100) ;
    line=get_lines_from_file(tmpname, &nline);
    
    for (i=0; i<nline; i++){
        fprintf(fout_in, "%s", line[i]);
    }
    
    if (VERB==0) delete_file(tmpname);
    free_lines_from_file(line, nline); 
    
}

void write_item(FILE *fout, char *item)
    
{
    fprintf(fout,"%7s ", item);
    return;
}

void write_sf_unmerge(FILE *fout,  char *block_name, char *pdbid, int i, int *index, int id)
{
//    char comment[100]; i, block number starting from 0
    long  n, m, n1, nstart=0;
    

    fprintf(fout, "data_%s\n",block_name);

    fprintf(fout,"#\n_entry.id  '%s'\n", pdbid);
//    strcpy(comment, "");
    
//    write_diffrn(fout, sf.ndif, sf.dif_id,sf.dif_cid,sf.dif_detail,sf.temp, sf.treat, comment);
    
                 
    fprintf(fout,"#\n_diffrn.id  1\n");
    fprintf(fout,"_diffrn.crystal_id  1\n");
    if (strlen(DETAIL)>1){
        fprintf(fout,"_diffrn.details  \n;%s\n;",  DETAIL);
    }else if (sf.dif_detail){
        fprintf(fout,"_diffrn.details  \n;%s\n;",  sf.dif_detail[0]);
    }
    if (strlen(DETAIL)>1 && sf.dif_detail){
        sprintf(INFO, "Warning: your given detail overwrited the existing detail.\n");
        pinfo(INFO,0);
    }
    
         
    fprintf(fout,"\n#\n_exptl_crystal.id  1\n");

    fprintf(fout,"#\n_diffrn_radiation.diffrn_id  1 \n");
    fprintf(fout,"_diffrn_radiation.wavelength_id  1 \n");
    
    if(sf.a &&  sf.b &&  sf.c && sf.alpha && sf.beta && sf.gamma){  
        fprintf(fout, "#\n_cell.entry_id  %s\n", pdbid);
        fprintf(fout, "_cell.length_a  %s\n", sf.a);
        fprintf(fout, "_cell.length_b  %s\n", sf.b);
        fprintf(fout, "_cell.length_c  %s\n", sf.c);
        fprintf(fout, "_cell.angle_alpha  %s\n", sf.alpha);
        fprintf(fout, "_cell.angle_beta  %s\n", sf.beta);
        fprintf(fout, "_cell.angle_gamma  %s\n", sf.gamma);
    }

    if(sf.sg) {
        fprintf(fout, "#\n_symmetry.entry_id  '%s'\n",pdbid);
        fprintf(fout, "_symmetry.space_group_name_H-M  '%s'\n",sf.sg);
    }
    if(sf.nsymm){
        fprintf(fout, "#\nloop_\n");
        fprintf(fout, "_symmetry_equiv.id\n");
        fprintf(fout, "_symmetry_equiv.pos_as_xyz\n");
        for(i=0; i<sf.nsymm; i++){
            fprintf(fout,"%d  '%s'\n", i+1,  sf.symm[i]);
        }
    }
    
    write_wavelength(fout, pdbid, i+1);
/*    
    if(sf.nwave && WAVE_CGI){
        fprintf(fout, "#\nloop_\n");
        fprintf(fout, "_diffrn_radiation_wavelength.id \n");
        fprintf(fout, "_diffrn_radiation_wavelength.wavelength \n");
        for(n=0; n<sf.nwave; n++){
            fprintf(fout,"%ld  %s \n", n+1,  sf.wavelength[n]);
        }
    }
*/    
    
//    nstart=index[i];
//    n1=index[i+1];

    nstart=0;
    n1=sf.nref;
    if (n1==0) n1=sf.dnref;
    

    if(n1>10){ // never change the order of conditions.
        write_unmerged_refl_head(fout);
    }

    m=0;
    for(n=nstart; n<n1; n++){
        m++;
        fprintf(fout, " 1 ");
        (sf.wave_id)? fprintf(fout, "%s", sf.wave_id[n]):fprintf(fout, "1");
        
        fprintf(fout, " 1 1 %ld ", m);

        (sf.H)? fprintf(fout,"%4s ", sf.H[n]) : fprintf(fout,"%4s ", sf.dH[n]) ;
        (sf.K)? fprintf(fout,"%4s ", sf.K[n]) : fprintf(fout,"%4s ", sf.dK[n]) ;
        (sf.L)? fprintf(fout,"%4s ", sf.L[n]) : fprintf(fout,"%4s ", sf.dL[n]) ;

        if(sf.Io){
            write_item(fout,sf.Io[n]);
            (sf.sIo)? fprintf(fout, "%s ", sf.sIo[n]):fprintf(fout, "? ");
        }else if (sf.F2o){
            write_item(fout,sf.F2o[n]);
            (sf.sF2o)? fprintf(fout, "%s ", sf.sF2o[n]):fprintf(fout, "? ");
        }else if (sf.I_plus){
            write_item(fout,sf.I_plus[n]);
            (sf.sI_plus)? fprintf(fout, "%s ", sf.sI_plus[n]):fprintf(fout, "? ");
        }else if (sf.I_minus){
            write_item(fout,sf.I_minus[n]);
            (sf.sI_minus)? fprintf(fout, "%s ", sf.sI_minus[n]):fprintf(fout, "? ");
        }else if (sf.unmerge_i){
            write_item(fout,sf.unmerge_i[n]);
            (sf.unmerge_si)? fprintf(fout, "%s ", sf.unmerge_si[n]):fprintf(fout, "? ");
        }else{
            fprintf(fout, "?  ? ");
        }
        
        fprintf(fout,"\n");
        
    } 
}

void write_unmerged_refl_head(FILE *fout)
{
    fprintf(fout,"#\nloop_\n");
    fprintf(fout,"_diffrn_refln.diffrn_id\n");
    fprintf(fout,"_diffrn_refln.wavelength_id\n");
    fprintf(fout,"_diffrn_refln.standard_code\n");
    fprintf(fout,"_diffrn_refln.scale_group_code\n");
    fprintf(fout,"_diffrn_refln.id\n");

    fprintf(fout,"_diffrn_refln.index_h\n");
    fprintf(fout,"_diffrn_refln.index_k\n");
    fprintf(fout,"_diffrn_refln.index_l\n");
    fprintf(fout,"_diffrn_refln.intensity_net\n");
    fprintf(fout,"_diffrn_refln.intensity_sigma\n");
}

int test_0_1(int start, int end, char **flag) 
{
    int nflag=0, k,  nn0=0, nn1=0;
    
    if (flag){//test 0 or 1 
        for (k=start; k<end; k++){
            if (atoi(flag[k])==0){
                nn0++;
            }else if (atoi(flag[k])==1){
                nn1++;
            }
            if (k-start >5000) break;
        }
        if (nn0>2*nn1) nflag=1;
    }
    return nflag;
}

void write_sf_status(FILE *fout, long n, int nflag)
{
    char flag=' ';
    

    if(TEST>0){ //dealing with status
        (rand()%TEST == 1)? flag='f':flag='o';
        fprintf(fout,"%c ",flag);
            
    }else if(sf.status){
        fprintf(fout,"%1s ",sf.status[n]);
                
    }else{
        if((sf.Fo_au && sf.Fo_au[n][0]!='?') ||  (sf.Io    && sf.Io[n][0]   !='?') ||
           (sf.I_plus &&sf.I_plus[n][0]!='?') || (sf.I_minus && sf.I_minus[n][0]!='?') ||
           (sf.F_plus &&sf.F_plus[n][0]!='?') || (sf.F_minus && sf.F_minus[n][0]!='?') ){
            if (sf.flag){
                if (strlen(FREERV)>0){
                    (atoi(FREERV)==atoi(sf.flag[n])) ? fprintf(fout," f ") : fprintf(fout," o ");
                }else{
                    (atoi(sf.flag[n])==nflag) ? fprintf(fout," f ") : fprintf(fout," o ");
                }
            }else if (sf.free1 && sf.free2){ // from phenix (can not re-assign FREERV)
                if (atoi(sf.free1[n])==nflag || atoi(sf.free2[n])==nflag){
                    fprintf(fout," f ");
                }else{
                    fprintf(fout," o ");
                }
            }else{
                fprintf(fout," o ");
            }
                
        }else{
            fprintf(fout," x ");
        }   
    }
}


void write_sf_file(FILE *fout,  char *block_name, int key, int i, int *index, int id)
/*
write all SF content

i, the datablock number
key==1 write all in each data_block in one file;
key==2 parse each separate set in each data_block & write them in each file
id==1 add Fo/sigFo if missing; id==0 do not add them
*/
    
{
    int nflag=0;
    long n, n1, nstart=0, m=0;
    float F=0, sF=0;
    
    if(key==2){ //each single set
        nstart=index[i];
        n1=index[i+1];
    }else{  // all sets in the block
        nstart=0;
        n1=sf.nref;
    }
    

///////////////
    if (sf.npow ){ //write powder diff
        
        fprintf(fout,"#\nloop_\n");

        if(sf.pow_sc) fprintf(fout,"_pdbx_powder_data.pd_meas_2theta_scan\n");
        if(sf.pow_io) fprintf(fout,"_pdbx_powder_data.pd_meas_intensity_total\n");
        if(sf.pow_iop)fprintf(fout,"_pdbx_powder_data.pd_proc_intensity_total\n");
        if(sf.pow_wt) fprintf(fout,"_pdbx_powder_data.pd_proc_ls_weight\n");
        if(sf.pow_ic) fprintf(fout,"_pdbx_powder_data.pd_proc_intensity_bkg_calc\n");
        if(sf.pow_ict)fprintf(fout,"_pdbx_powder_data.pd_calc_intensity_total\n");

        for(n=nstart; n<sf.npow; n++){
            if(sf.pow_sc)fprintf(fout,"%7s ", sf.pow_sc[n]);
            if(sf.pow_io)fprintf(fout,"%7s ", sf.pow_io[n]);
            if(sf.pow_iop)fprintf(fout,"%7s ", sf.pow_iop[n]);
            if(sf.pow_wt)fprintf(fout,"%7s ", sf.pow_wt[n]);
            if(sf.pow_ic)fprintf(fout,"%7s ", sf.pow_ic[n]);
            if(sf.pow_ict)fprintf(fout,"%7s ", sf.pow_ict[n]);
            fprintf(fout,"\n");
        }
        
        fprintf(fout,"\n#END\n");
    }
    
    if (sf.dnref){
        write_unmerged_refl_head(fout);
        
        for(n=nstart; n<sf.dnref; n++){
            fprintf(fout,"1 ");
            
            (sf.wave_id)? fprintf(fout, "%s", sf.wave_id[n]):fprintf(fout, "1");
            
            fprintf(fout," 1 1 %ld  %s %s %s  %s ", m++,
                    sf.dH[n],sf.dK[n],sf.dL[n], sf.unmerge_i[n]);
            (sf.unmerge_si)? fprintf(fout, " %s ",sf.unmerge_si[n]) : fprintf(fout, " ? ");
            fprintf(fout,"\n");
        }
        
        fprintf(fout,"\n#END\n");
    }
    
        
    if(n1>10){ // never change the order of conditions.

        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_refln.crystal_id\n");
        fprintf(fout,"_refln.wavelength_id\n");
        fprintf(fout,"_refln.scale_group_code\n");

        if(sf.H)fprintf(fout,"_refln.index_h\n");
        if(sf.K)fprintf(fout,"_refln.index_k\n");
        if(sf.L)fprintf(fout,"_refln.index_l\n");

        fprintf(fout,"_refln.status\n");
        if(sf.flag)fprintf(fout,"_refln.pdbx_r_free_flag\n");
        
        if(sf.Fo_au || sf.Fo)fprintf(fout,"_refln.F_meas_au\n"); 
        if(sf.sFo_au || sf.sFo)fprintf(fout,"_refln.F_meas_sigma_au\n");
        if(sf.Ic)fprintf(fout,"_refln.intensity_calc\n");
        if(sf.Io)fprintf(fout,"_refln.intensity_meas\n");
        if(sf.sIo)fprintf(fout,"_refln.intensity_sigma\n");
        
        if(sf.Fc)fprintf(fout,"_refln.F_calc\n");
        if(sf.Fc_au)fprintf(fout,"_refln.F_calc_au\n");
        if(sf.F2o)fprintf(fout,"_refln.F_squared_meas\n");
        if(sf.F2c)fprintf(fout,"_refln.F_squared_calc\n");
        if(sf.sF2o)fprintf(fout,"_refln.F_squared_sigma\n");
        
        if(sf.phase_c)fprintf(fout,"_refln.phase_calc\n");
        if(sf.phase_o)fprintf(fout,"_refln.phase_meas\n");
        
        if(sf.I_plus)fprintf(fout,"_refln.pdbx_I_plus\n");
        if(sf.sI_plus)fprintf(fout,"_refln.pdbx_I_plus_sigma\n");
        if(sf.I_minus)fprintf(fout,"_refln.pdbx_I_minus\n");
        if(sf.sI_minus)fprintf(fout,"_refln.pdbx_I_minus_sigma\n");
        
        if(sf.F_plus)fprintf(fout,"_refln.pdbx_F_plus\n");
        if(sf.sF_plus)fprintf(fout,"_refln.pdbx_F_plus_sigma\n");
        if(sf.F_minus)fprintf(fout,"_refln.pdbx_F_minus\n");
        if(sf.sF_minus)fprintf(fout,"_refln.pdbx_F_minus_sigma\n");
        
        if(sf.hla || sf.hla1)fprintf(fout,"_refln.pdbx_HL_A_iso\n");
        if(sf.hlb || sf.hlb1)fprintf(fout,"_refln.pdbx_HL_B_iso\n");
        if(sf.hlc || sf.hlc1)fprintf(fout,"_refln.pdbx_HL_C_iso\n");
        if(sf.hld || sf.hld1)fprintf(fout,"_refln.pdbx_HL_D_iso\n");

        if(sf.anom)fprintf(fout,"_refln.pdbx_anom_difference\n");
        if(sf.anoms)fprintf(fout,"_refln.pdbx_anom_difference_sigma\n");

        if(id==0 && !sf.Fo_au && !sf.Fo && (sf.I_plus || sf.F_plus || sf.Io)){
            fprintf(fout,"_refln.F_meas_au\n"); 
            fprintf(fout,"_refln.F_meas_sigma_au\n"); 
        }

        if(sf.d_spacing)fprintf(fout,"_refln.d_spacing\n");
        if(sf.a_calc)fprintf(fout,"_refln.A_calc\n");
        if(sf.b_calc)fprintf(fout,"_refln.B_calc\n");
//        if(sf.weighted)fprintf(fout,"_refln.pdbx_FWT\n");
        if(sf.fs_uncorrected)fprintf(fout,"_refln.F_meas_sigma_uncorrected\n");
        if(sf.f_uncorrected)fprintf(fout,"_refln.F_meas_uncorrected\n");
        if(sf.gsas_i100_meas)fprintf(fout,"_refln.pdbx_gsas_i100_meas\n");
        if(sf.i_unknown)fprintf(fout,"_refln.intensity_meas_unknown\n");
        if(sf.is_unknown)fprintf(fout,"_refln.intensity_sigma_unknown\n");

        if(sf.phase_part)fprintf(fout,"_refln.pdbx_phase_calc_part_solvent\n");
        if(sf.F_part_au) fprintf(fout,"_refln.pdbx_F_calc_part_solvent\n");
        if(sf.Fc_all)fprintf(fout,"_refln.pdbx_F_calc_with_solvent\n");
        if(sf.Fc_all_ph)fprintf(fout,"_refln.pdbx_phase_calc_with_solvent\n");

        if(sf.fwt)fprintf(fout,"_refln.pdbx_FWT\n");
        if(sf.phwt)fprintf(fout,"_refln.pdbx_PHWT\n");
        if(sf.delfwt)fprintf(fout,"_refln.pdbx_DELFWT\n");
        if(sf.delphwt)fprintf(fout,"_refln.pdbx_DELPHWT\n");
        
        if(sf.fom)fprintf(fout,"_refln.fom\n");
        
        if(sf.fib_co)fprintf(fout,"_refln.pdbx_fiber_coordinate\n");
        if(sf.fib_fo)fprintf(fout,"_refln.pdbx_fiber_F_meas_au\n");
        if(sf.fib_ly)fprintf(fout,"_refln.pdbx_fiber_layer\n");
        if(sf.wt)fprintf(fout,"_refln.weight\n");

    }
    
    if (sf.flag){        
        nflag=test_0_1(nstart, n1, sf.flag);
    }else if (sf.free1 && sf.free2){
        nflag=test_0_1(nstart, n1, sf.free1);
    }

    //printf("statust= %d|%s|\n", nflag,FREERV);

    for(n=nstart; n<n1; n++){
        
        (sf.cryst_id)? fprintf(fout, "%s ", sf.cryst_id[n]):fprintf(fout, "1 ");
        (sf.wave_id)? fprintf(fout, "%s ", sf.wave_id[n]):fprintf(fout, "1 ");
        (sf.scale_id)? fprintf(fout, "%s ", sf.scale_id[n]):fprintf(fout, "1 ");

        if(sf.H) fprintf(fout,"%4s ", sf.H[n]);
        if(sf.K) fprintf(fout,"%4s ", sf.K[n]);
        if(sf.L) fprintf(fout,"%4s ", sf.L[n]);

        write_sf_status(fout, n, nflag);
        
        
        if(sf.flag) fprintf(fout,"%2d ", atoi(sf.flag[n]));
        
        if(sf.Fo_au) {
            write_item(fout, sf.Fo_au[n]);
        }else if (sf.Fo) {
            write_item(fout, sf.Fo[n]);
        }
        
        if(sf.sFo_au){
            write_item(fout, sf.sFo_au[n]);
        }else if(sf.sFo){
            write_item(fout, sf.sFo[n]);   
        }
        
        if(sf.Ic)write_item(fout,sf.Ic[n]);
        if(sf.Io)write_item(fout,sf.Io[n]);
        if(sf.sIo)write_item(fout, sf.sIo[n]);
        
        if(sf.Fc)write_item(fout,sf.Fc[n]);        
        if(sf.Fc_au)write_item(fout,sf.Fc_au[n]);
        
        if(sf.F2o)write_item(fout,sf.F2o[n]);
        if(sf.F2c)write_item(fout,sf.F2c[n]);
        if(sf.sF2o)write_item(fout,sf.sF2o[n]);
        
        if(sf.phase_c) write_item(fout,sf.phase_c[n]);
        if(sf.phase_o) write_item(fout,sf.phase_o[n]);
        
        if(sf.I_plus) write_item(fout,sf.I_plus[n]);
        if(sf.sI_plus) write_item(fout,sf.sI_plus[n]);
        if(sf.I_minus) write_item(fout,sf.I_minus[n]);
        if(sf.sI_minus) write_item(fout,sf.sI_minus[n]);
       
        if(sf.F_plus) write_item(fout,sf.F_plus[n]);
        if(sf.sF_plus) write_item(fout,sf.sF_plus[n]);
        if(sf.F_minus) write_item(fout,sf.F_minus[n]);
        if(sf.sF_minus) write_item(fout, sf.sF_minus[n]);
       
        
        if(sf.hla){
            fprintf(fout,"%6s ", sf.hla[n]);
        }else if(sf.hla1){
            fprintf(fout,"%6s ", sf.hla1[n]);
        }
        
        if(sf.hlb){
            fprintf(fout,"%6s ", sf.hlb[n]);
        }else if(sf.hlb1){
            fprintf(fout,"%6s ", sf.hlb1[n]);
        }
        
        if(sf.hlc){
            fprintf(fout,"%6s ", sf.hlc[n]);
        }else if(sf.hlc1){
            fprintf(fout,"%6s ", sf.hlc1[n]);
        }
        
        if(sf.hld){
            fprintf(fout,"%6s ", sf.hld[n]);
        }else if(sf.hld1){
            fprintf(fout,"%6s ", sf.hld1[n]);
        }
        
        if(sf.anom)fprintf(fout,"%6s ", sf.anom[n]);
        if(sf.anoms)fprintf(fout,"%6s ", sf.anoms[n]);
        

        if(id==0 && !sf.Fo_au && !sf.Fo && (sf.I_plus || sf.F_plus || sf.Io)){
            other_to_f(n, F, sF) ;
            fprintf(fout," %7.2f %6.2f ",F, sF);
            
        }
        
        if(sf.d_spacing)fprintf(fout,"%3s ", sf.d_spacing[n]);
        if(sf.a_calc)fprintf(fout,"%7s ", sf.a_calc[n]);
        if(sf.b_calc)fprintf(fout,"%7s ", sf.b_calc[n]);
   //     if(sf.weighted)fprintf(fout,"%7s ", sf.weighted[n]);
        if(sf.fs_uncorrected)fprintf(fout,"%7s ", sf.fs_uncorrected[n]);
        if(sf.f_uncorrected)fprintf(fout,"%7s ", sf.f_uncorrected[n]);
        if(sf.gsas_i100_meas)fprintf(fout,"%7s ", sf.gsas_i100_meas[n]);
        if(sf.i_unknown)fprintf(fout,"%7s ", sf.i_unknown[n]);
        if(sf.is_unknown)fprintf(fout,"%7s ", sf.is_unknown[n]);

        if(sf.phase_part)fprintf(fout,"%7s ", sf.phase_part[n]);
        if(sf.F_part_au)fprintf(fout,"%7s ", sf.F_part_au[n]);
        
        if(sf.Fc_all)fprintf(fout,"%7s ", sf.Fc_all[n]);
        if(sf.Fc_all_ph)fprintf(fout,"%7s ", sf.Fc_all_ph[n]);
        
        if(sf.fwt)fprintf(fout,"%7s ", sf.fwt[n]);
        if(sf.phwt)fprintf(fout,"%7s ", sf.phwt[n]);
        if(sf.delfwt)fprintf(fout,"%7s ", sf.delfwt[n]);
        if(sf.delphwt)fprintf(fout,"%7s ", sf.delphwt[n]);
            
        if(sf.fom)fprintf(fout,"%5s ", sf.fom[n]);

        if(sf.fib_co)fprintf(fout,"%7s ", sf.fib_co[n]);
        if(sf.fib_fo)fprintf(fout,"%7s ", sf.fib_fo[n]);
        if(sf.fib_ly)fprintf(fout,"%7s ", sf.fib_ly[n]);
        if(sf.wt)fprintf(fout,"%7s ", sf.wt[n]);
           
        fprintf(fout,"\n");
    }
}

void check_sf(int j, float *resol)
// check various errors in the single set; j, the block number
{

    char hkl_min[80], hkl_max[80], hkl[80], mess[20], file[20];
    
    int ah,ak,al,key=0,nstart=0,n1=0,n2=0, n4=0,n5=0,n6=0,n7=0,n8=0;
    int max_H=-500, min_H=500, max_K=-500, min_K=500,max_L=-500, min_L=500,nblock;
    
    long nfp=0, nfn=0, nip=0, nin=0, n_obs=0, n_free=0,nnii=0, nnii_low=0, nref=0;
    long  n, i=0, nf_sFo=0, nf_sIo=0, nfpairF=0, nfpairI=0, nf_Fo=0, nf_F2o=0, nf_Io=0;
    
    float f, sigf,  rcell[6], resolution=0, rfree_p, cell[6];
    float max_F=-500000, min_F=500000, max_I=-5000000, min_I=5000000;
    float max_R=-500, min_R=900, max_F2=-500000, min_F2=500000;
    
    double sum_i=0, sum_si=0, sum_f=0, sum_sf=0, i_over_si=0, f_over_sf=0;
    double sum_f2=0, sum_sf2=0, f2_over_sf2=0;
    double val, sval, ii_sigii=0, ii_sigii_low=0,sum_sigii=0, sum_ii=0;
    double ratio, ii_sigii_max=-90000;
    
    nblock=j+1;
    *resol=100;
    
    nstart=0;
    n=sf.nref;
    strcpy(file, PDBID);

    check_cell(cell, j); //get cell for block j
    
    strcpy(mess,"Warning:");
    if (j==0) strcpy(mess,"Error:");
    
    if(! (sf.H || sf.K || sf.L || sf.dH || sf.dK || sf.dL )) {
        sprintf(INFO, "%s File (%s) has no 'index_h, index_k, index_l' (data block=%d).", mess,file, nblock);
        pinfo(INFO,0);
        return;
    }
    
    if (!(sf.Fo_au || sf.Fo || sf.Io || sf.F2o || sf.I_plus || sf.I_minus ||
          sf.F_plus || sf.F_minus || sf.unmerge_i || sf.unmerge_si)){
        sprintf(INFO, "%s File (%s) has no mandatory items 'F/I/F+/F-/I+/I-' (data block=%d). ", mess, file, nblock);
        pinfo(INFO,0);
        return;
    }
    
    if(n<30 && sf.dnref>30) return ; //do not further check the unmerged data!!
     
    
    if(n<30){
        sprintf(INFO, "Error:  File (%s) has too few reflections (%ld) (data block=%d).", file, n, nblock);
        pinfo(INFO,0);
        return;
    }
    
    if ((sf.Fo_au && !sf.sFo_au) || (sf.Fo && !sf.sFo)  ||
        (sf.Io && !sf.sIo) || (sf.F2o && !sf.sF2o) ||
        (sf.I_plus && !sf.sI_plus) || (sf.I_minus && !sf.sI_minus) ||
        (sf.F_plus && !sf.sF_plus) || (sf.F_minus && !sf.sF_minus) ||
        (sf.unmerge_i && !sf.unmerge_si)

        ){
        sprintf(INFO, "Warning: File (%s) sigma values are missing (data block=%d)!", file, nblock);
        pinfo(INFO,0);
    }
    
    if (j==0 && !sf.status){
        sprintf(INFO,"Warning: File (%s) has no free set (data block=%d).", file,nblock);
        pinfo(INFO, 0);
    }
    
    if(cell[0]>0.01 && cell[1]>0.01){
        key=1;
        calc_recip_cell(cell, rcell);
    }
    
    for(i=nstart; i<n; i++){ //check data items
        ah = atoi(sf.H[i]);
        ak = atoi(sf.K[i]);
        al = atoi(sf.L[i]);

        if(ah<min_H) min_H=ah;
        if(ak<min_K) min_K=ak;
        if(al<min_L) min_L=al;
       
        if(ah>max_H) max_H=ah;
        if(ak>max_K) max_K=ak;
        if(al>max_L) max_L=al;

        if (!( (sf.Fo_au && sf.Fo_au[i][0] != '?') || (sf.Io && sf.Io[i][0] != '?') ||
             (sf.I_plus && sf.I_plus[i][0] != '?') || (sf.I_minus && sf.I_minus[i][0] != '?') ||
             (sf.F_plus && sf.F_plus[i][0] != '?') || (sf.F_minus && sf.F_minus[i][0] != '?') ||
             (sf.unmerge_i && sf.unmerge_i[i][0] != '?')  ||
               (sf.F2o && sf.F2o[i][0] != '?')  || (sf.Fo && sf.Fo[i][0] != '?') )) {
//            printf("No values for H,K,L=%d  %d  %d \n", ah, ak,al);
            continue;
        }
        nref++;
        
        sprintf(hkl, "HKL=%4d %4d %4d", ah,ak,al);

        if((ah == 0 && ak == 0 && al == 0) ||
           (abs(ah) > 800 || abs(ak) > 800 || abs(al) > 800)){
            if (++n1 ==1){
                sprintf(INFO,"Error: File (%s) has wrong indices (%s).", file,hkl);
                pinfo(INFO,0);
            }
        }
        
        if (sf.sFo_au && i>0 && atof(sf.sFo_au[i-1])== atof(sf.sFo_au[i])) nf_sFo++;
        if (sf.sIo && i>0 && atof(sf.sIo[i-1])== atof(sf.sIo[i])) nf_sIo++;

        if (sf.F_plus){
            f=atof(sf.F_plus[i]);
            nfpairF++;
            if (f<0 && n4++ ==1){
                sprintf(INFO,"Error: File (%s) has negative amplitude (F+: %s) for (%s).",
                        file, sf.F_plus[i], hkl);
                pinfo(INFO,0);
            }
        }
        
        if (sf.I_plus){
            nfpairI++;
            f=atof(sf.I_plus[i]);
        }

        
        if(key>0 ){
            resolution = get_resolution(ah, ak, al, rcell);
            if(resolution<min_R){
                min_R=resolution;
                sprintf(hkl_min, "%4d %4d %4d", ah, ak, al);
            }
            if(resolution>max_R){
                max_R=resolution;
                sprintf(hkl_max, "%4d %4d %4d", ah, ak, al);
            }
            
            val=0;
            sval=0;
            if (sf.Io && sf.sIo && sf.sIo[i][0] != '?'){
                val=atof(sf.Io[i]);
                sval=atof(sf.sIo[i]);
            }else if ((sf.Fo_au && sf.sFo_au && sf.sFo_au[i][0] != '?') ){
                val=atof(sf.Fo_au[i])*atof(sf.Fo_au[i]);
                sval=2*atof(sf.Fo_au[i])*atof(sf.sFo_au[i]);
                
            }else if ((sf.F2o && sf.sFo && sf.sF2o[i][0] != '?') ){
                val=atof(sf.F2o[i]);
                sval=atof(sf.sF2o[i]);
            }else if ((sf.Fo && sf.sFo && sf.sFo[i][0] != '?') ){
                val=atof(sf.Fo[i])*atof(sf.Fo[i]);
                sval=2*atof(sf.Fo[i])*atof(sf.sFo[i]);
            }
        
            if (sval>0){
                sum_sigii= sum_sigii + sval; 
                sum_ii = sum_ii + val ;
                ratio = val/sval ;
                ii_sigii = ii_sigii + ratio;
                if (resolution>7.0){ //low resolution I/sigI
                    ii_sigii_low=ii_sigii_low + ratio;
                    nnii_low++;
                }
                if (ratio > ii_sigii_max) ii_sigii_max=ratio;
                
                nnii++;
            }
        }

        if (resolution <=RESOL +0.01 && resolution >=RESOH -0.01){ //for onedep
            if (sf.F_plus &&  !strchr(sf.F_plus[i], '?')  ) nfp++;
            if (sf.F_minus && !strchr(sf.F_minus[i], '?')) nfn++;
            if (sf.I_plus &&  !strchr(sf.I_plus[i], '?')) nip++;
            if (sf.I_minus && !strchr(sf.I_minus[i], '?')) nin++;

            if ((sf.Fo && !strchr(sf.Fo[i], '?')) || (sf.Fo_au && !strchr(sf.Fo_au[i], '?')) ||
                (sf.Io && !strchr(sf.Io[i], '?')) || (sf.F2o && !strchr(sf.F2o[i], '?'))||
                ((sf.F_plus && !strchr(sf.F_plus[i], '?')) || (sf.F_minus && !strchr(sf.F_minus[i], '?')))||
                ((sf.I_plus && !strchr(sf.I_plus[i], '?')) || (sf.I_minus && !strchr(sf.I_minus[i], '?')))){
                if (sf.status && strchr(sf.status[i],'o')){
                    n_obs++;
                }else if (sf.status && strchr(sf.status[i],'f')){
                    n_free++;
                }
            }
        }

//        printf("the resolution %.2f %.2f %ld\n", RESOH, RESOL, i);
        
        if(sf.Fo_au) {
            f=atof(sf.Fo_au[i]);
            if (f<0 && n5++ ==1){
                sprintf(INFO,"Error: File (%s) has negative amplitude (Fo: %s) for (%s).", file, sf.Fo_au[i], hkl);
                pinfo(INFO,0);
            }
            
            if(f<min_F) min_F=f;
            if(f>max_F) max_F=f;
            
            if(sf.sFo_au){
                sigf=atof(sf.sFo_au[i]);
                if(sigf>0 && sf.sFo_au[i][0] != '?'){
                    f_over_sf = f_over_sf + f/sigf;
                    sum_f = sum_f + f;
                    sum_sf = sum_sf + sigf;
                    nf_Fo++;
                }
            }
            
        }

        if(sf.F2o) {
            f=atof(sf.F2o[i]);
            if(f<min_F2) min_F2=f;
            if(f>max_F2) max_F2=f;
            if(sf.sF2o){
                sigf=atof(sf.sF2o[i]);
                if(sigf>0 && sf.sF2o[i][0] != '?'){
                    f2_over_sf2 = f2_over_sf2 + f/sigf;
                    sum_f2 = sum_f2 + f;
                    sum_sf2 = sum_sf2 + sigf;
                    nf_F2o++;
                }
            }
        }
        
        if(sf.Io) {
            f=atof(sf.Io[i]);
            if(f<min_I) min_I=f;
            if(f>max_I) max_I=f;
            if(sf.sIo){
                sigf=atof(sf.sIo[i]);
                if(sigf>0 && sf.sIo[i][0] != '?'){
                    i_over_si = i_over_si + f/sigf;
                    sum_i = sum_i + f;
                    sum_si = sum_si + sigf;
                    nf_Io++;
                    
                }
            }
        }
        
        if(sf.fom  && fabs(atof(sf.fom[i])) > 1.01) {
            if(n6++==1){
                sprintf(INFO,"Warning: File (%s) has wrong values of FOM (%s) for (%s).", file, sf.fom[i], hkl);
                pinfo(INFO, 0);
           }
       }
       if(sf.phase_c && fabs(atof(sf.phase_c[i]))>361.0){
            if(n7++==1){
                sprintf(INFO,"Warning: File (%s) has wrong values of phase (%s) for (%s).", file, sf.phase_c[i], hkl);
                pinfo(INFO, 0);
            }
       }
       if(sf.phase_o && fabs(atof(sf.phase_o[i]))>361.0){
            if(n8++==1){
                sprintf(INFO,"Warning: File (%s) has wrong values of phase (%s) for (%s).", file, sf.phase_o[i], hkl);
                pinfo(INFO, 0);
            }
        }

    }

    if(n1>0){
        sprintf(INFO,"Error: File (%s) has (%d) reflections with wrong indices.", file,n1);
        pinfo(INFO,0);
    }

            
    if (n2>0){
        sprintf(INFO,"Warning: File (%s) has (%d) reflections with negative SIGMA,(Corrected: given status '<'). ", file,n2);
        pinfo(INFO,0);
    }

    if (n4>0){
        sprintf(INFO,"Error: File (%s) has (%d) reflections with negative amplitude (F+). ", file,n4);
        pinfo(INFO,0);
    }
    
    if (n5>0){
        sprintf(INFO,"Error: File (%s) has (%d) reflections with negative amplitude (Fo). ", file,n5);
        pinfo(INFO,0);
    }
    

    if(nref>10 && ((nf_sFo>0 && nref-nf_sFo<3) || (nf_sIo>0 && nref-nf_sIo<3))) {
        sprintf(INFO,"Warning!  File (%s) has Sigma_Fo all the same!", file);
        pinfo(INFO,0);
    }
    
    sprintf(INFO,"Total number of observed reflections = %ld", (n_obs+n_free));
    pinfo(INFO, 1);

    sprintf(INFO,"Total number of observed reflections (status='o') = %ld", (n_obs));
    pinfo(INFO, 1);
    
    sprintf(INFO,"Total number of observed reflections (status='f') = %ld", (n_free));
    pinfo(INFO, 1);

    if (n_obs>0) {
        rfree_p=100*float(n_free)/float(n_obs);
        sprintf(INFO,"Percentage for free set = %6.2f", rfree_p);
        pinfo(INFO, 1);
    }
    

    if(nfpairF>10){
        sprintf(INFO, "Total number of Friedel pairs (F+/F-) = %ld", nfpairF);
        pinfo(INFO, 1);
        sprintf(INFO, "Total number of observed F+ = %ld", nfp );
        pinfo(INFO, 1);
        sprintf(INFO, "Total number of observed F- = %ld", nfn );
        pinfo(INFO, 1);
        sprintf(INFO, "Sum of observed F+ and F-  = %ld", (nfn+ nfp));
        pinfo(INFO, 1);
    }
    if(nfpairI>10){
        sprintf(INFO, "Total number of Friedel pairs (I+/I-) = %ld", nfpairI);
        pinfo(INFO, 1);
        sprintf(INFO, "Total number of observed I+ = %ld", nip );
        pinfo(INFO, 1);
        sprintf(INFO, "Total number of observed I- = %ld", nin );
        pinfo(INFO, 1);
        sprintf(INFO, "Sum of observed I+ and I-  = %ld", (nin+nip));
        pinfo(INFO, 1);
        
    }

    if(key>0 && cell[0]>0.001){
        
        sprintf(INFO,"Cell = %.2f %.2f %.2f %.2f %.2f %.2f", cell[0],cell[1],cell[2], cell[3],cell[4],cell[5]);
        pinfo(INFO, 1);
        
        sprintf(INFO, "Lowest resolution= %7.2f ; corresponding HKL=%s",max_R,hkl_max);
        pinfo(INFO, 1);
        sprintf(INFO, "Highest resolution=%7.2f ; corresponding HKL=%s",min_R,hkl_min);
        pinfo(INFO, 1);
        if (RESOH>0.11 && fabs(RESOH-min_R)>0.4 && nblock==1){
            sprintf(INFO, "Warning: large difference between reported (%.2f) and calculated(%.2f) resolution.\n ",RESOH,min_R);
            pinfo(INFO, 1);
        }
        *resol=min_R;
    }
    

    sprintf(INFO, "Max indices (Hmax=%4d  Kmax=%4d  Lmax=%4d) ",max_H, max_K, max_L);
    pinfo(INFO, 1);
    sprintf(INFO, "Min indices (Hmin=%4d  Kmin=%4d  Lmin=%4d) ",min_H, min_K, min_L);
    pinfo(INFO, 1);
    
    if(sf.Fo_au){
        sprintf(INFO,"maximum value of amplitude= %.2f",max_F);
        pinfo(INFO, 1);
        
        sprintf(INFO,"minimum value of amplitude= %.2f",min_F);
        pinfo(INFO, 1);

        sprintf(INFO,"<F/sigmaF> = %.2lf;  <F>/<sigmaF> = %.2lf; ",  f_over_sf/nf_Fo, sum_f/sum_sf );
        pinfo(INFO, 1);
        if (sum_f/sum_sf>140 || sum_f/sum_sf<4) {
            sprintf(INFO,"Warning: Value of (Fo_avg/sigFo_avg = %.2lf) is out of range (check Fo or SigFo in SF file). ",sum_f/sum_sf );
            pinfo(INFO, 0);
        }
        
    }
    
    if(sf.F2o){
        sprintf(INFO,"maximum value of F square= %.2f",max_F2);
        pinfo(INFO, 1);
        sprintf(INFO,"minimum value of F square= %.2f",min_F2);
        pinfo(INFO, 1);
        sprintf(INFO,"<F2/sigmaF2> = %.2lf;  <F2>/<sigmaF2> = %.2lf; ", f2_over_sf2/nf_F2o, sum_f2/sum_sf2 );
        pinfo(INFO, 1);
        
    }
    
    if(sf.Io){
        sprintf(INFO,"maximum value of intensity= %.2f",max_I);
        pinfo(INFO, 1);
        sprintf(INFO,"minimum value of intensity= %.2f",min_I);
        pinfo(INFO, 1);
        sprintf(INFO,"<I/sigmaI> = %.2lf;  <I>/<sigmaI> = %.2lf; ",i_over_si/nf_Io, sum_i/sum_si );
        pinfo(INFO, 1);
        if (sum_i/sum_si>80 || sum_i/sum_si<2) {
            sprintf(INFO,"Warning: Value of (I_avg/sigI_avg = %.2lf) is out of range (check Io or SigIo in SF file). ",sum_i/sum_si );
            pinfo(INFO, 0);
        }

        if (f_over_sf/nf_Fo>0 && (f_over_sf/nf_Fo > 2.5*i_over_si/nf_Io || f_over_sf/nf_Fo < 1.0*i_over_si/nf_Io )){
            sprintf(INFO,"Warning: too much difference Fo/sigFo = %.2lf;  Io/sigIo = %.2lf; ",f_over_sf/nf_Fo, i_over_si/nf_Io );
            pinfo(INFO, 0);
        }
        
    }

    if(nnii>10){
        sprintf(INFO,"Using all data:  <I/sigI>=%.2lf",ii_sigii/nnii);
        pinfo(INFO, 1);
        
        sprintf(INFO,"Using all data:  <I>/<sigI>=%.2lf", sum_ii/sum_sigii);
        pinfo(INFO, 1);
        
        sprintf(INFO,"Using all data:  Rsig(<sigI>/<I>)=%.3lf;",sum_sigii/sum_ii);
        pinfo(INFO, 1);
        
        sprintf(INFO,"The maximum value of <I/sigI>_max=%.2lf",ii_sigii_max );
        pinfo(INFO, 1);
    }
    
    if(nnii_low>10){
        sprintf(INFO,"Use data with resolution >7.0 Angstrom: <I/sigI>_low=%.2f",ii_sigii_low/nnii_low );
        pinfo(INFO, 1);
        
    }
    
    return ;
}

void check_sf_status(char **status, int j, char *file)
// do various status corrections
{
    
    char info[200];
    int  i, n0=0, nref=0, n_obs=0, n_free=0, id=-1, n1=0,n2=0,n3=0, nx=0, nh=0;
    int n, nmin=100, nmax=-100, nmin_1=0, nmax_1=0,nn0=0, nn1=0 ,nn2=0,  no=0, block;
    float f;

    block=j+1;
    
    for(i=0; i<sf.nref; i++){ //check status
//        printf ("stat=%4d %4d |%s\n", i, block, status[i]);
        
        if (status[i][0] == '.' ) {
            strcpy(status[i], "x");
            n1++;
        }else if (status[i][0]=='O' || status[i][0]=='F'){
            n2++;
        }else if (status[i][0]=='?'){
            n3++;
        }
        
        lowerstr(status[i]);
        
        if(status[i][0]=='x' ){
            nx++;
            continue;
        }else if (status[i][0]=='<' || status[i][0]=='h' || status[i][0]=='l'||
                  (status[i][0]=='-' && strlen(status[i])==1)) {
            nh++;
            continue;
        }
        
        if(status[i][0]=='o' ){
            n_obs++;
        }else if (status[i][0]=='f'){
            n_free++;
        }else {
            if (!isdigit(status[i][0])){
                no++;
                continue;
            }
            n=atoi(status[i]);
            if (n==0)nn0++;
            if (n==1)nn1++;
           
            if (n>=nmax) nmax=n;
            if (n<=nmin) nmin=n;
        }
            
        nref++;
    }

    for(i=0; i<sf.nref; i++){ //count numbers for nmin,nmax
        if (!isdigit(status[i][0])) continue;
        n=atoi(status[i]);
        if (n==nmin){
            nmin_1++;
        }else if (n==nmax) {
            nmax_1++;
        }
    }
    
    if (n1>0){
        sprintf(info,"Warning: (data block=%d): %d reflections have status assigned as '.' (changed to 'x')", block, n1);
        pinfo(info,0);
    }
    
    if (n2>0){
        sprintf(info,"Warning: (data block=%d): %d reflections have status assigned as 'F/O' (changed to 'f/o')", block,n2);
        pinfo(info,0);
    }
    
    if (n3>0){
        sprintf(info,"Warning: (data block=%d): %d reflections have status assigned as '?' (changed to 'x')", block, n3);
        pinfo(info,0);
    }
    
    if (no>0){
        sprintf(info,"Warning: (data block=%d): %d reflections have wrong status. Not in list(o,f,x,-,<,h,l)",block, no);
        pinfo(info,0);
    }
    
    
    
    if( nref==0 && nx > sf.nref -3 ){
        sprintf(info,"Warning: (data block=%d): all status (%d) are assigned as 'x' (changed! 'x'->'o')", block,nx);
        pinfo(info,0);
        id=1;
            
    }else if( n_free>50 &&  n_obs < n_free && nmin_1==0 && nmax_1==0 ){ //no numbers
        sprintf(info,"Warning: (data block=%d): status of free set('f') > observed('o') (corrected!).",block);
        pinfo(info,0);
        id=2;
            
    }else if(nmin_1>0 || nmax_1>0) { //with number
        if(n_free>10 ){  // a number assigned to 'o' but not 'f'
            sprintf(info,"Warning: (data block=%d): status 'o' is assigned as number '%d - %d'. (corrected! number->'o').",block,nmin, nmax);
            pinfo(info,0);
            id=3;
                
        }else if(n_obs>100){ // a number assigned to 'f' but not 'o'
            sprintf(info,"Warning: (data block=%d): status 'o' is assigned as number '%d - %d'. (corrected! number->'f').",block,nmin, nmax);
            pinfo(info,0);
            id=4;
            
        }else { //all numbers

            if( nmin == nmax && nmin !=100) { //the first one is  assigned as 'o'
                sprintf(info,"Warning: (data block=%d): all status assigned as number (%d~%d) (corrected! %d->'o').",block,nmin, nmax, nmin);
                id=7;
            }else if (nmin_1 > 3*nmax_1 && nmax_1>10 ){ //nmax is assigned as 'f'
                sprintf(info,"Warning: (data block=%d): status assigned as number (%d~%d) (corrected! %d->'f').",block,nmin, nmax, nmax);
                id=5;
            }else if ((nmax_1 > 3*nmin_1 && nmin_1>10) || (nmin_1>10 && nmax_1>10 && nmin_1<2*nmax_1)){ //nmin is assigned as 'f'
                if (strlen(FREERV)>0){
                    n0=atoi(FREERV);
                    sprintf(info,"Warning: (data block=%d): status assigned as number (%d~%d) (corrected! %d->'f').",block,nmin, nmax, n0);
                }else {
                    sprintf(info,"Warning: (data block=%d): status assigned as number (%d~%d) (corrected! %d->'f').",block,nmin, nmax, nmin);
                }
                id=6;
            }else{
                if(nn0>3*nn1 && nn1>10){
                    sprintf(info,"Warning: (data block=%d): status assigned as number (%d~%d) (corrected! 1->'f', check).",block,nmin, nmax);
                    id=8;
                }else if (nn1>3*nn0 && nn0>10 ){
                    sprintf(info,"Warning: (data block=%d): status assigned as number (%d~%d) (corrected! 0->'f', check).",block,nmin, nmax);
                    id=9;
                }
                
            }
            
            pinfo(info,0);
            
        }
    }
    
   // printf("FREERV, id= %s %d\n", FREERV, id);
      
    for(i=0; i<sf.nref; i++){  //do corrections

        
        if (status[i][0] == '?'){
            strcpy(status[i], "x") ;
        }else if (strlen(FREERV)>0 && id==6){ // manual input free R set (Given status)
            if(atoi(status[i])==atoi(FREERV)){
                strcpy(status[i], "f");
            }else
                strcpy(status[i], "o");
            
        }else if (id==1 ){ // all to 'o'
            if (status[i][0] == 'x') strcpy(status[i], "o") ;
            
        }else if (id==2){ //free set more than working set, swap
            if(strstr(status[i], "f"))
               strcpy(status[i], "o");
            else if(strstr(status[i], "o"))
                strcpy(status[i], "f");
            
        }else if (id==3){ // a number is assigned as 'o'
            if ((status[i][0]<=57 && status[i][0]>=48) ||
                (status[i][0] == '-' && strlen(status[i])>1 ) ) strcpy(status[i], "o");

        }else if (id==4){ // a number is assigned as 'f'
            if ((status[i][0]<=57 && status[i][0]>=48) ||
                (status[i][0] == '-' && strlen(status[i])>1 ) )  strcpy(status[i], "f");
            continue;
        }else if (id==5){ // nmax for free and others for 'o'
            (atoi(status[i]) == nmax) ? strcpy(status[i], "f"): strcpy(status[i], "o");
                
        }else if (id==6){ // nmin for free and others for 'o'
            (atoi(status[i])== nmin)? strcpy(status[i], "f"): strcpy(status[i], "o");
            
        }else if (id==7){ // 'o'= 'f', first one for 'o'
            (atoi(status[i])== nmin)? strcpy(status[i], "o"): strcpy(status[i], "f");
            
        }else if (id==8){ // 1 -> 'f'
            (atoi(status[i])== 1)? strcpy(status[i], "f"): strcpy(status[i], "o");
            
        }else if (id==9){ // 0 -> 'f'
            (atoi(status[i])== 0)? strcpy(status[i], "f"): strcpy(status[i], "o");
        }    
        if((sf.sFo_au && atof(sf.sFo_au[i])<0.0) || (sf.sIo && atof(sf.sIo[i])<0.0) ||
           (sf.Fo_au && sf.sFo_au && atof(sf.Fo_au[i])==-1.0 && atof(sf.sFo_au[i])==0.0) ){ //mainly for XDS
            if (sf.status && (sf.status[i][0]=='o' or sf.status[i][0]=='f')){
                sf.status[i][0]='<';
                nn2++;
            }
        }    
        if (no>0 && (strcmp(status[i], "o") && strcmp(status[i], "f")
                  && strcmp(status[i], "x")  && strcmp(status[i], "-")
                  && strcmp(status[i], "h")  && strcmp(status[i], "l")
                  && strcmp(status[i], "<"))) { // not converted
            sprintf(info,"Warning: (data block=%d): reflection status '%s' is converted to 'x'!", block,status[i]);
            strcpy(status[i], "x");
            pinfo(info,0);
            
        }
//        printf("status= %5d %s | %s \n", i,FREERV, status[i]);
        
    }
    if(nn2>0){
        sprintf(info,"Warning: (data block=%d): %d reflections having sigma<0 are assigned status as '<'! (unobserved)", block,nn2);
        pinfo(info,0);
    }

    if (nref==0) return;
    
    n0=0;
    n_obs=0;
    n_free=0;
    for(i=0; i<sf.nref; i++){ //final check for numbers or f/o swap
        if ((status[i][0]<=57 && status[i][0]>=48) ||
            (status[i][0] == '-' && strlen(status[i])>1 )) {
            n0++;
        }else if(status[i][0] =='o'){
            n_obs++;
        }else if(status[i][0] =='f'){
            n_free++;
        }
    }
    if (n0>0 && n_obs>0 && n_free>0){
        sprintf(info,"Warning: (data block=%d): status assigned as number (manual correction!).",block);
        pinfo(info,0);
        
    }else if(n_free>n_obs){
        sprintf(info,"Warning: (data block=%d): status more 'f' than 'o' (manual correction!).",block);
        pinfo(info,0);
        
    }
        
    if(block==0 && n_free<1 ){
        sprintf(info, "Error: SF file is missing R-free flags.\n");
        pinfo(info, 0);
    }
        
    if(n_obs>0 && n_free>0 ){
        f=100*float(n_free)/(n_obs+n_free);
        
        if(f>20){
            sprintf(info,"Warning: (data block=%d): too many reflections for the free set(nfree=%d, %.2f%%).",block,n_free, f);
            pinfo(info,0);
        }else if((f<1 && n_free<1000) || (f<6 && n_free<400) ){
            sprintf(info,"Warning: (data block=%d): too few reflections for the free set(nfree=%d, %.2f%%).",block,n_free, f);
            pinfo(info,0);
        }
            
    }
    
}


void correct_status(long n, char **status, int id)
{
    long i, nf=0;
    
    for(i=0; i<n; i++){
        if (id==0){ // Given status 
            if(atoi(status[i])==atoi(FREERV)){
                strcpy(status[i], "f");
                nf++;
            }else
                strcpy(status[i], "o");
            
        }else if (id==1){ // all to 'o'
            strcpy(status[i], "o");
            
        }else if (id==2){ //free set more than working set, swap
            if(strstr(status[i], "f"))
               strcpy(status[i], "o");
            else if(strstr(status[i], "o"))
                strcpy(status[i], "f");
            
        }else if (id==3){ // a number is assigned as 'o'
            if (status[i][0]<=57 && status[i][0]>=48) strcpy(status[i], "o");

        }else if (id==4){ // a number is assigned as 'f'
            if (status[i][0]<=57 && status[i][0]>=48) strcpy(status[i], "f");

            
        }else if (id==5){ // -1 for free and others for 'o'
            if(atoi(status[i])<0)
                strcpy(status[i], "f");
            else
                strcpy(status[i], "o");
                
        }else if (id==6){ // 1 for 'f' and 0 for 'o'
            if(atoi(status[i])==1)
                strcpy(status[i], "f");
            else
                strcpy(status[i], "o");
            
        }else if (id==7){ // 0 for 'o' and 1 for 'f'
            if(atoi(status[i])==0)
                strcpy(status[i], "o");
            else
                strcpy(status[i], "f");
            
        }
    }
        
}

        

float get_resolution(float h, float k, float l, float *rcell)
//calculate resolution based on HKL,rCELL.
{

    float aa1, aa2, aa3, a2, b2, c2, resol=0, dist_sq;
    const float DEG2RAD= 0.017453292;
    
    
    aa1 =  2 * rcell[0] * rcell[1] * cos(DEG2RAD*rcell[5]);
    aa2 =  2 * rcell[0] * rcell[2] * cos(DEG2RAD*rcell[4]);
    aa3 =  2 * rcell[1] * rcell[2] * cos(DEG2RAD*rcell[3]);

    a2 =  rcell[0]*rcell[0];
    b2 =  rcell[1]*rcell[1];
    c2 =  rcell[2]*rcell[2];

    dist_sq = h*h*a2 + k*k*b2 + l*l*c2 + h*k*aa1 + h*l*aa2 + k*l*aa3;
    if(dist_sq>0.0000001)
        resol = sqrt(1.0/dist_sq);

    return resol;
    
}
    
void calc_recip_cell(float *cell, float *rcell)
{
    float cosa,cosb, cosc,sina,sinb,sinc, cosast,cosbst,coscst,v;
    const float DEG2RAD= 0.017453292;
    
    
    cosa = cos(DEG2RAD*cell[3]);
    cosb = cos(DEG2RAD*cell[4]);
    cosc = cos(DEG2RAD*cell[5]);
    sina = sin(DEG2RAD*cell[3]);
    sinb = sin(DEG2RAD*cell[4]);
    sinc = sin(DEG2RAD*cell[5]);
    
    v = cell[0]* cell[1] *cell[2]*
        sqrt(1. - cosa*cosa - cosb*cosb - cosc*cosc + 2.*cosa*cosb*cosc );
    
    rcell[0] =  cell[1] * cell[2]* sina / v;
    rcell[1] =  cell[0] * cell[2]* sinb / v;
    rcell[2] =  cell[0] * cell[1]* sinc / v;
    
    cosast   =  (cosb*cosc-cosa) / (sinb*sinc);
    cosbst   =  (cosa*cosc-cosb) / (sina*sinc);
    coscst   =  (cosa*cosb-cosc) / (sina*sinb);
    
    rcell[3] =  acos(cosast)/DEG2RAD;
    rcell[4] =  acos(cosbst)/DEG2RAD;
    rcell[5] =  acos(coscst)/DEG2RAD;
    
}


void cif_token_change(const char *old_token, const char *new_token)
{
    printf("Warning! The mmcif token  _refln.%s is wrong!\n", old_token);
    printf("It has been corrected as _refln.%s \n",new_token);
}

void parse_sf_values(CifFile *fobjR, int i)
//parse all the sf values
{
    long n1=0, n2=0, n3=0, n4=0,  k1=0, k2=0, naud=0, ndif=0 ;
    string diags;
    const char *Bid;
    vector<string> blockNames;
    
    fobjR->GetBlockNames(blockNames);
    Bid = blockNames[i].c_str();
    
    // Empty values if we come back through here 2x in program
    sf.empty();

    sf.a = get_attribute_value(fobjR, Bid, "cell", "length_a");
    sf.b = get_attribute_value(fobjR, Bid, "cell", "length_b");
    sf.c = get_attribute_value(fobjR, Bid, "cell", "length_c");
    sf.alpha = get_attribute_value(fobjR, Bid, "cell", "angle_alpha");
    sf.beta = get_attribute_value(fobjR, Bid, "cell", "angle_beta");
    sf.gamma = get_attribute_value(fobjR, Bid, "cell", "angle_gamma");
    sf.sg = get_attribute_value(fobjR, Bid, "symmetry", "space_group_name_H-M");
    sf.nsg = get_attribute_value(fobjR, Bid, "symmetry", "Int_Tables_number");

    sf.resh=get_attribute_value(fobjR, Bid,"reflns", "d_resolution_high");
    if (!sf.resh) sf.resh=get_attribute_value(fobjR, Bid,"diffrn_reflns", "pdbx_d_res_high");
        
    sf.resl=get_attribute_value(fobjR, Bid,"reflns", "d_resolution_low");
    if (!sf.resl) sf.resl=get_attribute_value(fobjR, Bid,"diffrn_reflns", "pdbx_d_res_low");
    
    sf.hmax=get_attribute_value(fobjR, Bid,"reflns", "limit_h_max");
    if (!sf.hmax) sf.hmax=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_h_max");

    sf.kmax=get_attribute_value(fobjR, Bid,"reflns", "limit_k_max");
    if (!sf.kmax) sf.kmax=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_k_max");
    
    sf.lmax=get_attribute_value(fobjR, Bid,"reflns", "limit_l_max");
    if (!sf.lmax) sf.lmax=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_l_max");
    
    sf.hmin=get_attribute_value(fobjR, Bid,"reflns", "limit_h_min");
    if (!sf.hmin) sf.hmin=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_h_min");

    sf.kmin=get_attribute_value(fobjR, Bid,"reflns", "limit_k_min");
    if (!sf.kmin) sf.kmin=get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_k_min");

    sf.lmin=get_attribute_value(fobjR, Bid,"reflns", "limit_l_min");
    if (!sf.lmin)sf.lmin =get_attribute_value(fobjR, Bid,"diffrn_reflns", "limit_l_min");
    
    sf.nall=get_attribute_value(fobjR, Bid,"reflns", "number_all");
    if (!sf.nall) sf.nall=get_attribute_value(fobjR, Bid,"diffrn_reflns", "number");

    sf.nobs=get_attribute_value(fobjR, Bid,"reflns", "number_obs");
    if (!sf.nobs) sf.nobs=get_attribute_value(fobjR, Bid,"diffrn_reflns", "pdbx_number_obs");

    
    
    sf.wavelength = get_attribute_values(fobjR, Bid, "diffrn_radiation_wavelength", "wavelength", &k1);
    sf.symm=get_attribute_values(fobjR, Bid,"symmetry_equiv", "pos_as_xyz", &k2);

    sf.aud_id=get_attribute_values(fobjR, Bid,"audit", "revision_id", &naud);
    sf.aud_date=get_attribute_values(fobjR, Bid,"audit", "creation_date", &naud);
    sf.aud_rec=get_attribute_values(fobjR, Bid,"audit", "update_record", &naud);
    sf.aud_meth=get_attribute_values(fobjR, Bid,"audit", "creation_method", &naud);
    sf.dif_id=get_attribute_values(fobjR, Bid,"diffrn", "id", &ndif);
    sf.dif_cid=get_attribute_values(fobjR, Bid,"diffrn", "crystal_id", &ndif);
    sf.temp=get_attribute_values(fobjR, Bid,"diffrn", "ambient_temp", &ndif);
    sf.treat=get_attribute_values(fobjR, Bid,"diffrn", "crystal_treatment", &ndif);
    sf.dif_detail=get_attribute_values(fobjR, Bid,"diffrn", "details", &ndif);
    if(!sf.dif_detail) sf.dif_detail=get_attribute_values(fobjR, Bid,"diffrn", "detail", &ndif);
 
    sf.H = get_attribute_values(fobjR, Bid, "refln", "index_h", &n2);
    sf.K = get_attribute_values(fobjR, Bid, "refln", "index_k", &n2);
    sf.L = get_attribute_values(fobjR, Bid, "refln", "index_l", &n2);

    sf.Fo = get_attribute_values(fobjR, Bid, "refln", "F_meas", &n1);
    sf.sFo = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma", &n1);

    sf.Fo_au = get_attribute_values(fobjR, Bid, "refln", "F_meas_au", &n1);
    sf.sFo_au = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma_au", &n1);
    
    sf.Fc = get_attribute_values(fobjR, Bid, "refln", "F_calc", &n1);
    sf.Fc_au = get_attribute_values(fobjR, Bid, "refln", "F_calc_au", &n1);
    

    sf.Io = get_attribute_values(fobjR, Bid, "refln", "intensity_meas", &n1);
    if(!sf.Io){
        sf.Io = get_attribute_values(fobjR, Bid, "refln", "intensity_meas_au", &n1);
        if(sf.Io) cif_token_change("intensity_meas_au", "intensity_meas" );
        
        if(!sf.Io){
            sf.Io = get_attribute_values(fobjR, Bid, "refln", "intensity", &n1);
            if(sf.Io) cif_token_change("intensity", "intensity_meas" ); 
        }
    }
    
    
    sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma", &n1);
    if(!sf.sIo){
        sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma_au", &n1);
        if(sf.sIo) cif_token_change("intensity_sigma_au", "intensity_sigma" );

        if(!sf.sIo){
            sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_sigm", &n1);
            if(sf.sIo) cif_token_change("intensity_sigm", "intensity_sigma" );

            if(!sf.sIo){
                sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_meas_sigma", &n1);
                if(sf.sIo) cif_token_change("intensity_meas_sigma", "intensity_sigma" );

                if(!sf.sIo){
                    sf.sIo = get_attribute_values(fobjR, Bid, "refln", "intensity_meas_sigma_au", &n1);
                    if(sf.sIo) cif_token_change("intensity_meas_sigma_au", "intensity_sigma" );
                }
            }
            
        }

    }

    
    sf.F2o = get_attribute_values(fobjR, Bid, "refln", "F_squared_meas", &n1);
    sf.sF2o = get_attribute_values(fobjR, Bid, "refln", "F_squared_sigma", &n1);
    sf.Ic = get_attribute_values(fobjR, Bid, "refln", "intensity_calc", &n1);
    sf.F2c = get_attribute_values(fobjR, Bid, "refln", "F_squared_calc", &n1);
    
    sf.fom = get_attribute_values(fobjR, Bid, "refln", "fom", &n1);
    sf.phase_c = get_attribute_values(fobjR, Bid, "refln", "phase_calc", &n1);
    sf.phase_o = get_attribute_values(fobjR, Bid, "refln", "phase_meas", &n1);

    sf.status = get_attribute_values(fobjR, Bid, "refln", "status", &n1);
    if(!sf.status){
        sf.status = get_attribute_values(fobjR, Bid, "refln", "R_free_flag", &n1);
        if(sf.status) cif_token_change("R_free_flag", "status" );

        if(!sf.status){
            sf.status = get_attribute_values(fobjR, Bid, "refln", "statu", &n1);
            if(sf.status) cif_token_change("statu", "status" );

            if(!sf.status){
                sf.status = get_attribute_values(fobjR, Bid, "refln", "status_au", &n1);
                if(sf.status) cif_token_change("status_au", "status" );
            }
        }
    }
    sf.flag = get_attribute_values(fobjR, Bid, "refln", "pdbx_r_free_flag", &n1);
    
        //sf.free1 & sf.free2 special for two freeset, merge one to another if ?
    sf.free1 = get_attribute_values(fobjR, Bid, "refln", "R-free-flags(+)", &n1);
    sf.free2 = get_attribute_values(fobjR, Bid, "refln", "R-free-flags(-)", &n1);
    
        
    sf.wave_id = get_attribute_values(fobjR, Bid, "refln", "wavelength_id", &n1);
    sf.cryst_id = get_attribute_values(fobjR, Bid, "refln", "crystal_id", &n1);
    sf.scale_id = get_attribute_values(fobjR, Bid, "refln", "scale_group_code", &n1);
    
    sf.I_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus", &n1);
    sf.sI_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_plus_sigma", &n1);
    sf.I_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus", &n1);
    sf.sI_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_I_minus_sigma", &n1);
     
    sf.F_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_plus", &n1);
    sf.sF_plus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_plus_sigma", &n1);
    sf.F_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_minus", &n1);
    sf.sF_minus = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_minus_sigma", &n1);
    
    sf.hla = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_A_iso", &n1);
    sf.hlb = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_B_iso", &n1);
    sf.hlc = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_C_iso", &n1);
    sf.hld = get_attribute_values(fobjR, Bid, "refln", "pdbx_HL_D_iso", &n1);

    sf.hla1 = get_attribute_values(fobjR, Bid, "refln", "pdbx_HLA", &n1);
    sf.hlb1 = get_attribute_values(fobjR, Bid, "refln", "pdbx_HLB", &n1);
    sf.hlc1 = get_attribute_values(fobjR, Bid, "refln", "pdbx_HLC", &n1);
    sf.hld1 = get_attribute_values(fobjR, Bid, "refln", "pdbx_HLD", &n1);

    sf.anom =  get_attribute_values(fobjR, Bid, "refln", "pdbx_anom_difference", &n1);
    if(!sf.anom){
        sf.anom =  get_attribute_values(fobjR, Bid, "refln", "DF_anomalous", &n1);
        if(sf.anom) cif_token_change("DF_anomalous", "pdbx_anom_difference" ); 
    }
    
    sf.anoms = get_attribute_values(fobjR, Bid, "refln", "pdbx_anom_difference_sigma", &n1);
    if(!sf.anoms){
        sf.anoms =  get_attribute_values(fobjR, Bid, "refln", "DF_anomalous_sigma", &n1);
        if(sf.anoms) cif_token_change("DF_anomalous_sigma", "pdbx_anom_difference_sigma" ); 
    }

    sf.d_spacing = get_attribute_values(fobjR, Bid, "refln", "d_spacing", &n1);
    sf.a_calc = get_attribute_values(fobjR, Bid, "refln", "A_calc", &n1);
    sf.b_calc = get_attribute_values(fobjR, Bid, "refln", "B_calc", &n1);
    sf.weighted = get_attribute_values(fobjR, Bid, "refln", "pdbx_fom_weighted_fmap", &n1);
    sf.fs_uncorrected = get_attribute_values(fobjR, Bid, "refln", "F_meas_sigma_uncorrected", &n1);
    sf.f_uncorrected = get_attribute_values(fobjR, Bid, "refln", "F_meas_uncorrected", &n1);
    sf.gsas_i100_meas = get_attribute_values(fobjR, Bid, "refln", "pdbx_gsas_i100_meas", &n1);
    sf.i_unknown = get_attribute_values(fobjR, Bid, "refln", "intensity_meas_unknown", &n1);
    sf.is_unknown = get_attribute_values(fobjR, Bid, "refln", "intensity_sigma_unknown", &n1);
    sf.phase_part= get_attribute_values(fobjR, Bid, "refln", "phase_part", &n1);
    sf.F_part_au = get_attribute_values(fobjR, Bid, "refln", "F_part_au", &n1);

    sf.Fc_all = get_attribute_values(fobjR, Bid, "refln", "pdbx_F_calc_with_solvent", &n1);
    sf.Fc_all_ph = get_attribute_values(fobjR, Bid, "refln", "pdbx_phase_calc_with_solvent", &n1);
    sf.fwt = get_attribute_values(fobjR, Bid, "refln", "pdbx_FWT", &n1);
    sf.phwt = get_attribute_values(fobjR, Bid, "refln", "pdbx_PHWT", &n1);
    sf.delfwt = get_attribute_values(fobjR, Bid, "refln", "pdbx_DELFWT", &n1);
    sf.delphwt = get_attribute_values(fobjR, Bid, "refln", "pdbx_DELPHWT", &n1);

    sf.fib_co = get_attribute_values(fobjR, Bid, "refln", "pdbx_fiber_coordinate", &n1);
    sf.fib_fo = get_attribute_values(fobjR, Bid, "refln", "pdbx_fiber_F_meas_au", &n1);
    sf.fib_ly = get_attribute_values(fobjR, Bid, "refln", "pdbx_fiber_layer", &n1);

    sf.wt = get_attribute_values(fobjR, Bid, "refln", "weight", &n1);

    sf.pow_sc = get_attribute_values(fobjR, Bid, "pdbx_powder_data", "pd_meas_2theta_scan", &n3);
    sf.pow_io = get_attribute_values(fobjR, Bid, "pdbx_powder_data", "pd_meas_intensity_total", &n3);
    sf.pow_iop =get_attribute_values(fobjR, Bid, "pdbx_powder_data", "pd_proc_intensity_total", &n3);
    sf.pow_wt = get_attribute_values(fobjR, Bid, "pdbx_powder_data", "pd_proc_ls_weight", &n3);
    sf.pow_ic = get_attribute_values(fobjR, Bid, "pdbx_powder_data", "pd_proc_intensity_bkg_calc", &n3);
    sf.pow_ict =get_attribute_values(fobjR, Bid, "pdbx_powder_data", "pd_calc_intensity_total", &n3);

    sf.dH = get_attribute_values(fobjR, Bid, "diffrn_refln", "index_h", &n4);
    sf.dK = get_attribute_values(fobjR, Bid, "diffrn_refln", "index_k", &n4);
    sf.dL = get_attribute_values(fobjR, Bid, "diffrn_refln", "index_l", &n4);
    
    sf.unmerge_i = get_attribute_values(fobjR, Bid, "diffrn_refln", "intensity_net", &n4);
    sf.unmerge_si =get_attribute_values(fobjR, Bid, "diffrn_refln", "intensity_sigma", &n4);

//printf("n2=%ld\n", n2);
    sf.dnref=n4;
    sf.npow = n3;
    sf.nref = n2;
    sf.nwave = k1;
    sf.nsymm = k2;
    sf.naud = naud;
    sf.ndif = ndif;
    
                
}

